from flask import request
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app.services import (
    assets_service,
    breakdown_service,
    entities_service,
    projects_service,
    shots_service,
    user_service,
)

from zou.app.mixin import ArgsMixin


class CastingResource(Resource):
    @jwt_required()
    def get(self, project_id, entity_id):
        """
        Resource to retrieve the casting of a given entity.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: entity_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting of given entity
        """
        user_service.check_project_access(project_id)
        return breakdown_service.get_casting(entity_id)

    @jwt_required()
    def put(self, project_id, entity_id):
        """
        Resource to allow the modification of assets linked to an entity.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: entity_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Modification of assets linked to an entity
        """
        casting = request.json
        user_service.check_manager_project_access(project_id)
        return breakdown_service.update_casting(entity_id, casting)


class EpisodesCastingResource(Resource):
    @jwt_required()
    def get(self, project_id):
        """
        Resource to retrieve the casting of episodes.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting of episodes
        """
        user_service.check_project_access(project_id)
        return breakdown_service.get_production_episodes_casting(project_id)


class EpisodeSequenceAllCastingResource(Resource):
    @jwt_required()
    def get(self, project_id, episode_id):
        """
        Resource to retrieve the casting of shots from given episode.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: episode_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting for all shots from given episode.
        """
        user_service.check_project_access(project_id)
        return breakdown_service.get_all_sequences_casting(
            project_id, episode_id=episode_id
        )


class SequenceAllCastingResource(Resource):
    @jwt_required()
    def get(self, project_id):
        """
        Resource to retrieve the casting of shots from all sequences of given
        project.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting for all shots from given project.
        """
        user_service.check_project_access(project_id)
        return breakdown_service.get_all_sequences_casting(project_id)


class SequenceCastingResource(Resource):
    @jwt_required()
    def get(self, project_id, sequence_id):
        """
        Resource to retrieve the casting of shots from given sequence.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: sequence_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting of shots from given sequence
        """
        user_service.check_project_access(project_id)
        shots_service.get_sequence(sequence_id)
        return breakdown_service.get_sequence_casting(sequence_id)


class AssetTypeCastingResource(Resource):
    @jwt_required()
    def get(self, project_id, asset_type_id):
        """
        Resource to retrieve the casting of assets from given asset type.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: asset_type_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Casting of assets from given asset type
        """
        user_service.check_project_access(project_id)
        assets_service.get_asset_type(asset_type_id)
        return breakdown_service.get_asset_type_casting(
            project_id, asset_type_id
        )


class ShotAssetInstancesResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, shot_id):
        """
        Retrieve all asset instances linked to shot.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All assets linked to shot
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        return breakdown_service.get_asset_instances_for_shot(shot_id)

    @jwt_required()
    def post(self, shot_id):
        """
        Add an asset instance to given shot.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Asset instance added to given shot
        """
        args = self.get_args([("asset_instance_id", None, True)])

        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        shot = breakdown_service.add_asset_instance_to_shot(
            shot_id, args["asset_instance_id"]
        )
        return shot, 201


class RemoveShotAssetInstanceResource(Resource, ArgsMixin):
    @jwt_required()
    def delete(self, shot_id, asset_instance_id):
        """
        Remove an asset instance from given shot.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: shot_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: asset_instance_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Asset instance removed from given shot
        """
        shot = shots_service.get_shot(shot_id)
        user_service.check_project_access(shot["project_id"])
        shot = breakdown_service.remove_asset_instance_for_shot(
            shot_id, asset_instance_id
        )
        return "", 204


class SceneAssetInstancesResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, scene_id):
        """
        Retrieve all asset instances linked to scene.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All asset instances linked to given scene
        """
        scene = shots_service.get_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        return breakdown_service.get_asset_instances_for_scene(scene_id)

    @jwt_required()
    def post(self, scene_id):
        """
        Create an asset instance on given scene.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Asset instances created on given scene
        """
        args = self.get_args(
            [("asset_id", None, True), ("description", None, False)]
        )

        scene = shots_service.get_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        asset_instance = breakdown_service.add_asset_instance_to_scene(
            scene_id, args["asset_id"], args["description"]
        )
        return asset_instance, 201


class SceneCameraInstancesResource(Resource):
    @jwt_required()
    def get(self, scene_id):
        """
        Retrieve all camera instances linked to scene.
        ---
        tags:
          - Breakdown
        parameters:
          - in: path
            name: scene_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: All camera instances linked to scene
        """
        scene = shots_service.get_scene(scene_id)
        user_service.check_project_access(scene["project_id"])
        return breakdown_service.get_camera_instances_for_scene(scene_id)


class ProjectEntityLinksResource(Resource, ArgsMixin):
    @jwt_required()
    def get(self, project_id):
        """
        Retrieve all entity links related to given project.

        Results can be paginated using page and limit query parameters. If you
        prefer a more accurate pagination, you can use and
        cursor_created_at to get the next page.
        ---
        tags:
          - Breakdown
        description: It's mainly used for synchronisation purpose.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: page
            required: False
            type: string
            format: Number
            x-example: 2
          - in: query
            name: limit
            required: False
            type: string
            format: Number
            x-example: 100
          - in: query
            name: cursor_created_at
            required: False
            type: string
            format: Datetime
            x-example: 2020-01-01T00:00:00
        responses:
            200:
                description: All entity links related to given project
        """
        user_service.check_manager_project_access(project_id)
        projects_service.get_project(project_id)
        page = self.get_page()
        limit = self.get_limit()
        cursor_created_at = self.get_text_parameter("cursor_created_at")
        return entities_service.get_entity_links_for_project(
            project_id,
            page=page,
            limit=limit,
            cursor_created_at=cursor_created_at,
        )


class ProjectEntityLinkResource(Resource):
    @jwt_required()
    def delete(self, project_id, entity_link_id):
        """
        Delete given entity link.
        ---
        tags:
          - Breakdown
        description: It's mainly used for synchronisation purpose.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: entity_link_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Entity link deleted
        """
        user_service.check_manager_project_access(project_id)
        return entities_service.remove_entity_link(entity_link_id)



from flask import request, abort
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from sqlalchemy.exc import StatementError

from zou.app.models.data_import_error import DataImportError


class ShotgunImportErrorsResource(Resource):
    def __init__(self):
        Resource.__init__(self)

    @jwt_required()
    def get(self):
        """
        Import shotgun error resource.
        ---
        tags:
          - Import
        responses:
            200:
                description: Resource imported
        """
        criterions = {"source": "shotgun"}
        import_errors = DataImportError.query.filter_by(**criterions).all()
        return DataImportError.serialize_list(import_errors)

    @jwt_required()
    def post(self):
        """
        Serialize shotgun error resource.
        ---
        tags:
          - Import
        responses:
            200:
                description: Resource serialized
        """
        error = DataImportError(event_data=request.json, source="shotgun")
        error.save()
        return error.serialize(), 201


class ShotgunImportErrorResource(Resource):
    def __init__(self):
        Resource.__init__(self)

    @jwt_required()
    def delete(self, error_id):
        """
        Delete error.
        ---
        tags:
          - Import
        parameters:
          - in: path
            name: error_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Error deleted
            404:
                description: Error non-existant or Statement error
        """
        try:
            error = DataImportError.get(error_id)
        except StatementError:
            abort(404)

        if error is None:
            abort(404)
        error.delete()

        return {"deletion_success": True}, 204



from functools import wraps
from flask_principal import RoleNeed, Permission
from werkzeug.exceptions import Forbidden

admin_permission = Permission(RoleNeed("admin"))
manager_permission = Permission(RoleNeed("manager"))
supervisor_permission = Permission(RoleNeed("supervisor"))
client_permission = Permission(RoleNeed("client"))
vendor_permission = Permission(RoleNeed("vendor"))
artist_permission = Permission(RoleNeed("user"))

bot_permission = Permission(RoleNeed("bot"))
person_permission = Permission(RoleNeed("person"), RoleNeed("person_api"))
person_api_permission = Permission(RoleNeed("person_api"))


class PermissionDenied(Forbidden):
    pass


def has_manager_permissions():
    """
    Return True if user is an admin or a manager.
    """
    return admin_permission.can() or manager_permission.can()


def has_artist_permissions():
    """
    Return True if user is an artist.
    """
    return artist_permission.can()


def has_admin_permissions():
    """
    Return True if user is an admin.
    """
    return admin_permission.can()


def has_client_permissions():
    """
    Return True if user is a client.
    """
    return client_permission.can()


def has_vendor_permissions():
    """
    Return True if user is a vendor.
    """
    return vendor_permission.can()


def has_supervisor_permissions():
    """
    Return True if user is a supervisor.
    """
    return supervisor_permission.can()


def has_person_permissions():
    """
    Return True if user is a person.
    """
    return person_permission.can()


def has_at_least_supervisor_permissions():
    """
    Return True if user is an admin or a manager.
    """
    return (
        supervisor_permission.can()
        or admin_permission.can()
        or manager_permission.can()
    )


def check_at_least_supervisor_permissions():
    """
    Return True if user is admin, manager or supervsior. It raises a
    PermissionDenied exception in case of failure.
    """
    if has_at_least_supervisor_permissions():
        return True
    else:
        raise PermissionDenied


def check_manager_permissions():
    """
    Return True if user is admin or manager. It raises a PermissionDenied
    exception in case of failure.
    """
    if has_manager_permissions():
        return True
    else:
        raise PermissionDenied


def check_admin_permissions():
    """
    Return True if user is admin. It raises a PermissionDenied exception in case
    of failure.
    """
    if admin_permission.can():
        return True
    else:
        raise PermissionDenied


def check_person_permissions():
    """
    Return True if user is a person. It raises a PermissionDenied exception in
    case of failure.
    """
    if person_permission.can():
        return True
    else:
        raise PermissionDenied


def require_admin(function):
    @wraps(function)
    def decorated_function(*args, **kwargs):
        check_admin_permissions()
        return function(*args, **kwargs)

    return decorated_function


def require_manager(function):
    @wraps(function)
    def decorated_function(*args, **kwargs):
        check_manager_permissions()
        return function(*args, **kwargs)

    return decorated_function


def require_person(function):
    @wraps(function)
    def decorated_function(*args, **kwargs):
        check_person_permissions()
        return function(*args, **kwargs)

    return decorated_function



from flask import Blueprint
from flask_jwt_extended import jwt_required
from zou.app.utils.api import configure_api_from_blueprint
from zou.app.utils import permissions

from zou.app.blueprints.user.resources import (
    AssetTasksResource,
    AssetTaskTypesResource,
    AssetTypeAssetsResource,
    ClearAvatarResource,
    ContextResource,
    DateTimeSpentsResource,
    DayOffResource,
    ChatsResource,
    JoinChatResource,
    OpenProjectsResource,
    ProjectEpisodesResource,
    ProjectSequencesResource,
    ProjectAssetTypesResource,
    SceneTasksResource,
    SceneTaskTypesResource,
    SequenceTasksResource,
    SequenceTaskTypesResource,
    SequenceShotsResource,
    SequenceScenesResource,
    ShotTasksResource,
    ShotTaskTypesResource,
    ToChecksResource,
    TodosResource,
    DoneResource,
    FilterResource,
    FiltersResource,
    FilterGroupResource,
    FilterGroupsResource,
    DesktopLoginLogsResource,
    MarkAllNotificationsAsReadResource,
    NotificationsResource,
    NotificationResource,
    HasTaskSubscribedResource,
    TaskSubscribeResource,
    TaskUnsubscribeResource,
    TaskTimeSpentResource,
    TimeSpentsResource,
    HasSequenceSubscribedResource,
    SequenceSubscriptionsResource,
    SequenceSubscribeResource,
    SequenceUnsubscribeResource,
)

routes = [
    ("/data/user/context", ContextResource),
    ("/data/user/assets/<asset_id>/tasks", AssetTasksResource),
    ("/data/user/shots/<shot_id>/tasks", ShotTasksResource),
    ("/data/user/scenes/<scene_id>/tasks", SceneTasksResource),
    ("/data/user/sequences/<sequence_id>/tasks", SequenceTasksResource),
    ("/data/user/assets/<asset_id>/task-types", AssetTaskTypesResource),
    ("/data/user/shots/<shot_id>/task-types", ShotTaskTypesResource),
    ("/data/user/scenes/<scene_id>/task-types", SceneTaskTypesResource),
    (
        "/data/user/sequences/<sequence_id>/task-types",
        SequenceTaskTypesResource,
    ),
    ("/data/user/projects/open", OpenProjectsResource),
    (
        "/data/user/projects/<project_id>/asset-types",
        ProjectAssetTypesResource,
    ),
    (
        "/data/user/projects/<project_id>/asset-types/<asset_type_id>/assets",
        AssetTypeAssetsResource,
    ),
    ("/data/user/projects/<project_id>/sequences", ProjectSequencesResource),
    ("/data/user/projects/<project_id>/episodes", ProjectEpisodesResource),
    ("/data/user/sequences/<sequence_id>/shots", SequenceShotsResource),
    ("/data/user/sequences/<sequence_id>/scenes", SequenceScenesResource),
    ("/data/user/tasks", TodosResource),
    ("/data/user/tasks-to-check", ToChecksResource),
    ("/data/user/done-tasks", DoneResource),
    ("/data/user/filters", FiltersResource),
    ("/data/user/filters/<filter_id>", FilterResource),
    ("/data/user/filter-groups", FilterGroupsResource),
    ("/data/user/filter-groups/<filter_group_id>", FilterGroupResource),
    ("/data/user/desktop-login-logs", DesktopLoginLogsResource),
    ("/data/user/time-spents", TimeSpentsResource),
    ("/data/user/time-spents/<date>", DateTimeSpentsResource),
    ("/data/user/tasks/<task_id>/time-spents/<date>", TaskTimeSpentResource),
    ("/data/user/day-offs/<date>", DayOffResource),
    ("/data/user/notifications", NotificationsResource),
    ("/data/user/notifications/<notification_id>", NotificationResource),
    ("/data/user/tasks/<task_id>/subscribed", HasTaskSubscribedResource),
    ("/data/user/chats", ChatsResource),
    ("/actions/user/chats/<entity_id>/join", JoinChatResource),
    ("/actions/user/tasks/<task_id>/subscribe", TaskSubscribeResource),
    ("/actions/user/tasks/<task_id>/unsubscribe", TaskUnsubscribeResource),
    ("/actions/user/clear-avatar", ClearAvatarResource),
    (
        "/data/user/entities/<sequence_id>/task-types/<task_type_id>/subscribed",
        HasSequenceSubscribedResource,
    ),
    (
        "/data/user/projects/<project_id>/task-types/<task_type_id>/sequence-subscriptions",
        SequenceSubscriptionsResource,
    ),
    (
        "/actions/user/sequences/<sequence_id>/task-types/<task_type_id>/subscribe",
        SequenceSubscribeResource,
    ),
    (
        "/actions/user/sequences/<sequence_id>/task-types/<task_type_id>/unsubscribe",
        SequenceUnsubscribeResource,
    ),
    (
        "/actions/user/notifications/mark-all-as-read",
        MarkAllNotificationsAsReadResource,
    ),
]

blueprint = Blueprint("user", "user")
api = configure_api_from_blueprint(
    blueprint,
    routes,
    decorators=[
        jwt_required(),
        permissions.require_person,
    ],
)



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.files.resources import (
    WorkingFilePathResource,
    LastWorkingFilesResource,
    ModifiedFileResource,
    CommentWorkingFileResource,
    NewWorkingFileResource,
    TaskWorkingFilesResource,
    EntityWorkingFilesResource,
    EntityOutputFilePathResource,
    GetNextEntityOutputFileRevisionResource,
    NewEntityOutputFileResource,
    LastEntityOutputFilesResource,
    EntityOutputTypesResource,
    EntityOutputTypeOutputFilesResource,
    InstanceOutputFilePathResource,
    NewInstanceOutputFileResource,
    GetNextInstanceOutputFileRevisionResource,
    LastInstanceOutputFilesResource,
    InstanceOutputTypesResource,
    InstanceOutputTypeOutputFilesResource,
    EntityOutputFilesResource,
    ProjectOutputFilesResource,
    InstanceOutputFilesResource,
    SetTreeResource,
    FileResource,
    WorkingFileFileResource,
    GuessFromPathResource,
)

routes = [
    ("/data/files/<file_id>", FileResource),
    ("/data/tasks/<task_id>/working-files", TaskWorkingFilesResource),
    ("/data/tasks/<task_id>/working-files/new", NewWorkingFileResource),
    (
        "/data/tasks/<task_id>/working-files/last-revisions",
        LastWorkingFilesResource,
    ),
    ("/data/tasks/<task_id>/working-file-path", WorkingFilePathResource),
    (
        "/data/asset-instances/<asset_instance_id>"
        "/entities/<temporal_entity_id>/output-files/new",
        NewInstanceOutputFileResource,
    ),
    (
        "/data/asset-instances/<asset_instance_id>"
        "/entities/<temporal_entity_id>/output-files/next-revision",
        GetNextInstanceOutputFileRevisionResource,
    ),
    (
        "/data/asset-instances/<asset_instance_id>"
        "/entities/<temporal_entity_id>/output-files/last-revisions",
        LastInstanceOutputFilesResource,
    ),
    (
        "/data/asset-instances/<asset_instance_id>"
        "/entities/<temporal_entity_id>/output-types",
        InstanceOutputTypesResource,
    ),
    (
        "/data/asset-instances/<asset_instance_id>"
        "/entities/<temporal_entity_id>/output-types"
        "/<output_type_id>/output-files",
        InstanceOutputTypeOutputFilesResource,
    ),
    (
        "/data/asset-instances/<asset_instance_id>"
        "/entities/<temporal_entity_id>/output-file-path",
        InstanceOutputFilePathResource,
    ),
    ("/data/entities/<entity_id>/working-files", EntityWorkingFilesResource),
    (
        "/data/entities/<entity_id>/output-files/new",
        NewEntityOutputFileResource,
    ),
    (
        "/data/entities/<entity_id>/output-files/next-revision",
        GetNextEntityOutputFileRevisionResource,
    ),
    (
        "/data/entities/<entity_id>/output-files/last-revisions",
        LastEntityOutputFilesResource,
    ),
    ("/data/entities/<entity_id>/output-types", EntityOutputTypesResource),
    (
        "/data/entities/<entity_id>/output-types/<output_type_id>/output-files",
        EntityOutputTypeOutputFilesResource,
    ),
    ("/data/entities/<entity_id>/output-files", EntityOutputFilesResource),
    ("/data/projects/<project_id>/output-files", ProjectOutputFilesResource),
    (
        "/data/asset-instances/<asset_instance_id>/output-files",
        InstanceOutputFilesResource,
    ),
    (
        "/data/entities/<entity_id>/output-file-path",
        EntityOutputFilePathResource,
    ),
    ("/data/entities/guess_from_path", GuessFromPathResource),
    (
        "/data/working-files/<working_file_id>/file",
        WorkingFileFileResource,
    ),
    ("/actions/projects/<project_id>/set-file-tree", SetTreeResource),
    (
        "/actions/working-files/<working_file_id>/comment",
        CommentWorkingFileResource,
    ),
    (
        "/actions/working-files/<working_file_id>/modified",
        ModifiedFileResource,
    ),
]

blueprint = Blueprint("files", "files")
api = configure_api_from_blueprint(blueprint, routes)



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.entities.resources import (
    EntityPreviewFilesResource,
    EntityNewsResource,
    EntityTimeSpentsResource,
    EntitiesLinkedWithTasksResource,
)


routes = [
    ("/data/entities/<entity_id>/news", EntityNewsResource),
    ("/data/entities/<entity_id>/preview-files", EntityPreviewFilesResource),
    ("/data/entities/<entity_id>/time-spents", EntityTimeSpentsResource),
    (
        "/data/entities/<entity_id>/entities-linked/with-tasks",
        EntitiesLinkedWithTasksResource,
    ),
]

blueprint = Blueprint("entities", "entities")
api = configure_api_from_blueprint(blueprint, routes)



"""empty message

Revision ID: 556526e47daa
Revises: 6f6049877105
Create Date: 2019-03-23 22:01:57.576051

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "556526e47daa"
down_revision = "6f6049877105"
branch_labels = None
depends_on = None

ORIGINS = [("web", "Web"), ("script", "Script")]


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "login_log",
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column(
            "person_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "ip_address",
            sqlalchemy_utils.types.ip_address.IPAddressType(length=50),
            nullable=True,
        ),
        sa.Column(
            "origin",
            sqlalchemy_utils.types.choice.ChoiceType(ORIGINS),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["person_id"],
            ["person.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_login_log_person_id"),
        "login_log",
        ["person_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_login_log_created_at"),
        "login_log",
        ["created_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_login_log_person_id"), table_name="login_log")
    op.drop_index(op.f("ix_login_log_created_at"), table_name="login_log")
    op.drop_table("login_log")
    # ### end Alembic commands ###



import requests

from saml2 import (
    BINDING_HTTP_POST,
    BINDING_HTTP_REDIRECT,
)
from saml2.client import Saml2Client
from saml2.config import Config as Saml2Config

from zou.app import config


def saml_client_for(metadata_url):
    """
    Given the name of an IdP, return a configuation.
    The configuration is a hash for use by saml2.config.Config
    """
    acs_url = (
        f"{config.DOMAIN_PROTOCOL}://{config.DOMAIN_NAME}/api/auth/saml/sso"
    )

    rv = requests.get(metadata_url)

    settings = {
        "entityid": f"{config.DOMAIN_PROTOCOL}://{config.DOMAIN_NAME}/api/auth/saml/login",
        "metadata": {"inline": [rv.text]},
        "service": {
            "sp": {
                "endpoints": {
                    "assertion_consumer_service": [
                        (acs_url, BINDING_HTTP_REDIRECT),
                        (acs_url, BINDING_HTTP_POST),
                    ],
                },
                # Don't verify that the incoming requests originate from us via
                # the built-in cache for authn request ids in pysaml2
                "allow_unsolicited": True,
                # Don't sign authn requests, since signed requests only make
                # sense in a situation where you control both the SP and IdP
                "authn_requests_signed": False,
                "logout_requests_signed": True,
                "want_assertions_signed": True,
                "want_response_signed": False,
            },
        },
    }
    spConfig = Saml2Config()
    spConfig.load(settings)
    spConfig.allow_unknown_attributes = True
    saml_client = Saml2Client(config=spConfig)
    return saml_client



import os
import shutil
import time
from flask_fs.errors import FileNotFound

import errno


class DownloadFromStorageFailedException(Exception):
    pass


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def rm_rf(path):
    if os.path.exists(path):
        shutil.rmtree(path)


def rm_file(path):
    if os.path.exists(path):
        os.remove(path)


def copyfile(src, dest):
    shutil.copyfile(src, dest)


def get_file_path_and_file(
    config,
    get_local_path,
    open_file,
    prefix,
    instance_id,
    extension,
    file_size=None,
):
    if config.FS_BACKEND == "local":
        file_path = get_local_path(prefix, instance_id)
        if is_unvalid_file(file_path, file_size):
            raise FileNotFound
    else:
        file_path = os.path.join(
            config.TMP_DIR, "cache-%s-%s.%s" % (prefix, instance_id, extension)
        )

        if is_unvalid_file(file_path, file_size):
            download_failed = False
            exception = None
            try:
                with open(file_path, "wb") as tmp_file:
                    for chunk in open_file(prefix, instance_id):
                        tmp_file.write(chunk)
            except:
                download_failed = True

            if is_unvalid_file(
                file_path, file_size, download_failed
            ):  # download failed
                time.sleep(3)
                download_failed = False
                try:
                    with open(file_path, "wb") as tmp_file:
                        for chunk in open_file(prefix, instance_id):
                            tmp_file.write(chunk)
                except Exception as e:
                    download_failed = True
                    exception = e

                if is_unvalid_file(
                    file_path, file_size, download_failed
                ):  # download failed again
                    rm_file(file_path)
                    if exception is not None:
                        raise exception
                    else:
                        raise DownloadFromStorageFailedException

    return file_path


def is_unvalid_file(file_path, file_size=None, download_failed=False):
    """
    Check if file is absent, is empty or does match given size.
    """
    if download_failed or not os.path.exists(file_path):
        return True
    elif file_size is None:
        return get_file_size(file_path) == 0
    else:
        current_size = get_file_size(file_path)
        return current_size != file_size


def save_file(tmp_folder, instance_id, file_to_save):
    """
    Save file in given folder. The file must only be temporary saved via
    this function.
    """
    extension = "." + file_to_save.filename.split(".")[-1].lower()
    file_name = instance_id + extension.lower()
    file_path = os.path.join(tmp_folder, file_name)
    file_to_save.save(file_path)
    return file_path


def get_file_extension(filename):
    """
    Return extension of given file name in lower case.
    """
    return filename.split(".")[-1].lower()


def get_file_size(file_path):
    """
    Return in bytes the file size of the file located at given path.
    """
    return os.path.getsize(file_path)



"""empty message

Revision ID: a65bdadbae2f
Revises: e1ef93f40d3d
Create Date: 2019-01-15 12:19:59.813805

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
from sqlalchemy.dialects import postgresql
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "a65bdadbae2f"
down_revision = "e1ef93f40d3d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "metadata_descriptor",
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column(
            "project_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("entity_type", sa.String(length=60), nullable=False),
        sa.Column("name", sa.String(length=120), nullable=False),
        sa.Column("field_name", sa.String(length=120), nullable=False),
        sa.Column(
            "choices", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "project_id", "entity_type", "name", name="metadata_descriptor_uc"
        ),
    )
    op.create_index(
        op.f("ix_metadata_descriptor_entity_type"),
        "metadata_descriptor",
        ["entity_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_metadata_descriptor_project_id"),
        "metadata_descriptor",
        ["project_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_metadata_descriptor_project_id"),
        table_name="metadata_descriptor",
    )
    op.drop_index(
        op.f("ix_metadata_descriptor_entity_type"),
        table_name="metadata_descriptor",
    )
    op.drop_table("metadata_descriptor")
    # ### end Alembic commands ###



from slack import WebClient as SlackClient
from matterhook import Webhook
from discord import (
    Client as DiscordClient,
    Intents as DiscordIntents,
    Embed as DiscordEmbed,
)
from zou.app import config
import asyncio
import traceback
import logging

logger = logging.getLogger(__name__)
loghandler = logging.StreamHandler()
loghandler.setLevel(logging.INFO)
formatter = logging.Formatter(
    "%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
)
loghandler.setFormatter(formatter)
logger.addHandler(loghandler)


def send_to_slack(token, userid, message):
    if token:
        if userid:
            try:
                client = SlackClient(token=token)
                blocks = [
                    {
                        "type": "section",
                        "text": {"type": "mrkdwn", "text": message},
                    }
                ]
                client.chat_postMessage(
                    channel="@%s" % userid,
                    blocks=blocks,
                    as_user=True,
                )
            except Exception:
                logger.info("Exception when sending a Slack notification:")
                logger.info(traceback.format_exc())
        else:
            logger.info("The userid of Slack user is not defined.")
    else:
        logger.info(
            "The token of Slack for sending notifications is not defined."
        )


def send_to_mattermost(webhook, userid, message):
    if webhook:
        if userid:
            try:
                arg = webhook.split("/")
                server = "%s%s//%s" % (arg[0], arg[1], arg[2])
                hook = arg[4]

                # mandatory parameters are url and your webhook API key
                mwh = Webhook(server, hook)
                mwh.username = "Kitsu - %s" % (message["project_name"])
                mwh.icon_url = "%s://%s/img/kitsu.b07d6464.png" % (
                    config.DOMAIN_PROTOCOL,
                    config.DOMAIN_NAME,
                )

                # send a message to the API_KEY's channel
                mwh.send(message["message"], channel="@%s" % userid)

            except Exception:
                logger.info(
                    "Exception when sending a Mattermost notification:"
                )
                logger.info(traceback.format_exc())
        else:
            logger.info("The userid of Mattermost user is not defined.")
    else:
        logger.info(
            "The webhook of Mattermost for sending notifications is not defined."
        )


def send_to_discord(token, userid, message):
    async def send_to_discord_async(token, userid, message):
        intents = DiscordIntents.default()
        intents.members = True
        client = DiscordClient(intents=intents)

        @client.event
        async def on_ready(userid=userid, message=message):
            user_found = False
            for user in client.get_all_members():
                if not user.bot and (
                    (user.discriminator == "0" and user.name == userid)
                    or (f"{user.name}#{user.discriminator}" == userid)
                ):
                    embed = DiscordEmbed()
                    embed.description = message
                    await user.send(embed=embed)
                    user_found = True
                    break
            if not user_found:
                logger.info("User %s not found by Discord bot" % userid)

            await client.close()

        await client.start(token)

    if token:
        if userid:
            try:
                asyncio.set_event_loop(asyncio.new_event_loop())
                loop = asyncio.get_event_loop()
                loop.run_until_complete(
                    send_to_discord_async(token, userid, message)
                )
            except Exception:
                logger.info("Exception when sending a Discord notification:")
                logger.info(traceback.format_exc())
            finally:
                if loop:
                    loop.close()
        else:
            logger.info("The userid of Discord user is not defined.")
    else:
        logger.info(
            "The token of the Discord bot for sending notifications is not defined."
        )



from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class Organisation(db.Model, BaseMixin, SerializerMixin):
    """
    Model to represent current organisation settings.
    """

    name = db.Column(db.String(80), unique=True, nullable=False)
    hours_by_day = db.Column(db.Float, default=8, nullable=False)
    has_avatar = db.Column(db.Boolean(), default=False)
    use_original_file_name = db.Column(db.Boolean(), default=False)
    timesheets_locked = db.Column(db.Boolean(), default=False)
    format_duration_in_hours = db.Column(db.Boolean(), default=False)
    hd_by_default = db.Column(db.Boolean(), default=False)
    chat_token_slack = db.Column(db.String(80), default="")
    chat_webhook_mattermost = db.Column(db.String(80), default="")
    chat_token_discord = db.Column(db.String(80), default="")
    dark_theme_by_default = db.Column(db.Boolean(), default=False)
    format_duration_in_hours = db.Column(db.Boolean(), default=False)

    def present(self, sensitive=False):
        return self.serialize(
            ignored_attrs=(
                []
                if sensitive
                else [
                    "chat_token_slack",
                    "chat_webhook_mattermost",
                    "chat_token_discord",
                ]
            )
        )



from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class OutputType(db.Model, BaseMixin, SerializerMixin):
    """
    Type of an output files (geometry, cache, etc.)
    """

    name = db.Column(db.String(40), unique=True, nullable=False, index=True)
    short_name = db.Column(db.String(20), nullable=False)



"""empty message

Revision ID: de8a3de227ef
Revises: 45c2de366e66
Create Date: 2018-10-22 19:51:59.580865

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "de8a3de227ef"
down_revision = "0ef6416a507b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "comment_preview_link",
        sa.Column(
            "comment",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "preview_file",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["comment"],
            ["comment.id"],
        ),
        sa.ForeignKeyConstraint(
            ["preview_file"],
            ["preview_file.id"],
        ),
        sa.PrimaryKeyConstraint("comment", "preview_file"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("comment_preview_link")
    # ### end Alembic commands ###



"""create unique constraint project_id, field_name on MetadataDescriptor

Revision ID: 29df910f04a4
Revises: 8588f254d6b8
Create Date: 2022-11-04 13:46:15.819315

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy_utils import UUIDType, ChoiceType
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = "29df910f04a4"
down_revision = "8588f254d6b8"
branch_labels = None
depends_on = None

base = declarative_base()


class Department(base, BaseMixin):
    """
    Studio department like modeling, animation, etc.
    """

    __tablename__ = "department"
    name = sa.Column(sa.String(80), unique=True, nullable=False)
    color = sa.Column(sa.String(7), nullable=False)


department_metadata_descriptor_link = sa.Table(
    "department_metadata_descriptor_link",
    base.metadata,
    sa.Column(
        "metadata_descriptor_id",
        UUIDType(binary=False),
        sa.ForeignKey("metadata_descriptor.id"),
    ),
    sa.Column(
        "department_id", UUIDType(binary=False), sa.ForeignKey("department.id")
    ),
)


class MetadataDescriptor(base, BaseMixin):
    """
    This models allow to identify which metadata are available for a given
    project and a given entity type.
    """

    __tablename__ = "metadata_descriptor"

    project_id = sa.Column(
        UUIDType(binary=False),
        sa.ForeignKey("project.id"),
        nullable=False,
        index=True,
    )
    entity_type = sa.Column(sa.String(60), nullable=False, index=True)
    name = sa.Column(sa.String(120), nullable=False)
    field_name = sa.Column(sa.String(120), nullable=False)
    choices = sa.Column(postgresql.JSONB(astext_type=sa.Text()))
    for_client = sa.Column(sa.Boolean(), default=False, index=True)
    departments = orm.relationship(
        "Department", secondary=department_metadata_descriptor_link
    )

    __table_args__ = (
        sa.UniqueConstraint(
            "project_id", "entity_type", "name", name="metadata_descriptor_uc"
        ),
    )


PROJECT_STYLES = [
    ("2d", "2D Animation"),
    ("3d", "3D Animation"),
    ("2d3d", "2D/3D Animation"),
    ("vfx", "VFX"),
    ("stop-motion", "Stop Motion"),
    ("motion-design", "Motion Design"),
    ("archviz", "Archviz"),
    ("commercial", "Commercial"),
    ("catalog", "Catalog"),
]


class Project(base, BaseMixin):
    __tablename__ = "project"

    name = sa.Column(sa.String(80), nullable=False, unique=True, index=True)
    code = sa.Column(sa.String(80))
    description = sa.Column(sa.Text())
    shotgun_id = sa.Column(sa.Integer)
    file_tree = sa.Column(postgresql.JSONB(astext_type=sa.Text()))
    data = sa.Column(postgresql.JSONB(astext_type=sa.Text()))
    has_avatar = sa.Column(sa.Boolean(), default=False)
    fps = sa.Column(sa.String(10))
    ratio = sa.Column(sa.String(10))
    resolution = sa.Column(sa.String(12))
    production_type = sa.Column(sa.String(20), default="short")
    production_style = sa.Column(ChoiceType(PROJECT_STYLES))
    start_date = sa.Column(sa.Date())
    end_date = sa.Column(sa.Date())
    man_days = sa.Column(sa.Integer)
    nb_episodes = sa.Column(sa.Integer, default=0)
    episode_span = sa.Column(sa.Integer, default=0)
    max_retakes = sa.Column(sa.Integer, default=0)
    is_clients_isolated = sa.Column(sa.Boolean(), default=False)

    project_status_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("project_status.id"), index=True
    )


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    distinct_metadata_descriptors = session.query(
        MetadataDescriptor.entity_type,
        MetadataDescriptor.project_id,
        MetadataDescriptor.field_name,
    ).distinct()
    for metadata_descriptor in distinct_metadata_descriptors:
        metadata_descriptors_found = (
            session.query(MetadataDescriptor)
            .filter_by(
                entity_type=metadata_descriptor.entity_type,
                project_id=metadata_descriptor.project_id,
                field_name=metadata_descriptor.field_name,
            )
            .all()
        )
        for metadata_descriptor_to_remove in metadata_descriptors_found[1:]:
            session.delete(metadata_descriptor_to_remove)
    session.commit()
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "metadata_descriptor_uc2",
        "metadata_descriptor",
        ["project_id", "entity_type", "field_name"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "metadata_descriptor_uc2", "metadata_descriptor", type_="unique"
    )
    # ### end Alembic commands ###



import base64

import orjson as json
import os
import zlib

from flask import current_app
from operator import itemgetter
from pathlib import Path
from shutil import copyfile
from zipfile import ZipFile

from flask_fs.errors import FileNotFound
from slugify import slugify
from sqlalchemy import or_

from zou.app import config
from zou.app.stores import file_store

from zou.app.models.build_job import BuildJob
from zou.app.models.playlist import Playlist
from zou.app.models.preview_file import PreviewFile
from zou.app.models.task import Task
from zou.app.models.task_type import TaskType

from zou.utils import movie
from zou.app.utils import fields, events, remote_job, emails
from zou.app.utils import query as query_utils

from zou.app.services import (
    assets_service,
    base_service,
    entities_service,
    files_service,
    preview_files_service,
    projects_service,
    shots_service,
    tasks_service,
    names_service,
    persons_service,
)

from zou.app.services.exception import (
    BuildJobNotFoundException,
    PlaylistNotFoundException,
)


def all_playlists_for_project(
    project_id,
    for_client=False,
    page=1,
    sort_by="updated_at",
    task_type_id=None,
):
    """
    Return all playlists created for given project.
    """
    result = []
    query = Playlist.query.filter(Playlist.project_id == project_id)
    if for_client:
        query = query.filter(Playlist.for_client)

    if task_type_id is not None and len(task_type_id) > 0:
        query = query.filter(Playlist.task_type_id == task_type_id)

    query = query_utils.apply_sort_by(Playlist, query, sort_by)
    if page < 1:
        page = 1
    limit = 20
    offset = (page - 1) * limit
    query = query.limit(limit)
    query = query.offset(offset)
    playlists = query.all()
    for playlist in playlists:
        playlist_dict = build_playlist_dict(playlist)
        result.append(playlist_dict)
    return result


def all_playlists_for_episode(
    project_id,
    episode_id,
    for_client=False,
    page=1,
    sort_by="updated_at",
    task_type_id=None,
):
    """
    Return all playlists created for given episode.
    """
    result = []
    query = Playlist.query
    if for_client:
        query = query.filter(Playlist.for_client)

    if task_type_id is not None and len(task_type_id) > 0:
        query = query.filter(Playlist.task_type_id == task_type_id)

    if episode_id == "main":
        query = (
            query.filter(Playlist.episode_id == None)
            .filter(Playlist.project_id == project_id)
            .filter(
                or_(Playlist.is_for_all is None, Playlist.is_for_all == False)
            )
        )
    elif episode_id == "all":
        query = (
            query.filter(Playlist.episode_id == None)
            .filter(Playlist.project_id == project_id)
            .filter(Playlist.is_for_all == True)
        )
    else:
        query = query.filter(Playlist.episode_id == episode_id)

    query = query_utils.apply_sort_by(Playlist, query, sort_by)
    if page < 1:
        page = 1
    limit = 20
    offset = (page - 1) * limit
    query = query.limit(limit)
    query = query.offset(offset)
    playlists = query.all()
    for playlist in playlists:
        playlist_dict = build_playlist_dict(playlist)
        result.append(playlist_dict)
    return result


def build_playlist_dict(playlist):
    """
    Build a dictionary of a simplified version of the playlist. It just takes
    the information needed for displaying the list of playlists.
    """
    first_shot_preview_file_id = get_first_shot_preview_file_id(playlist)
    updated_at = fields.serialize_value(playlist.updated_at)
    playlist.shots = []
    playlist_dict = fields.serialize_value(playlist)
    del playlist_dict["shots"]
    playlist_dict["updated_at"] = updated_at
    if first_shot_preview_file_id is not None:
        playlist_dict["first_preview_file_id"] = first_shot_preview_file_id
    if playlist.for_entity is None:
        playlist_dict["for_entity"] = "shot"
    return playlist_dict


def get_first_shot_preview_file_id(playlist):
    first_shot_preview_file_id = None
    if (
        playlist.shots is not None
        and len(playlist.shots) > 0
        and isinstance(playlist.shots, list)
        and "preview_file_id" in playlist.shots[0]
    ):
        first_shot_preview_file_id = playlist.shots[0]["preview_file_id"]
    return first_shot_preview_file_id


def get_playlist_with_preview_file_revisions(playlist_id):
    """
    Return given playlist. Shot list is augmented with all previews available
    for a given shot.
    """
    playlist = get_playlist_raw(playlist_id)
    playlist_dict = playlist.serialize()
    playlist_dict = _add_build_job_infos_to_playlist_dict(
        playlist, playlist_dict
    )

    if playlist_dict["shots"] is None:
        playlist_dict["shots"] = []
    (playlist_dict, preview_file_map) = set_preview_files_for_entities(
        playlist_dict
    )

    for shot in playlist_dict["shots"]:
        try:
            preview_file = preview_file_map.get(shot["preview_file_id"], None)
            if preview_file is not None:
                shot["preview_file_id"] = preview_file["id"]
                shot["preview_file_extension"] = preview_file["extension"]
                shot["preview_file_revision"] = preview_file["revision"]
                shot["preview_file_width"] = preview_file["width"]
                shot["preview_file_height"] = preview_file["height"]
                shot["preview_file_duration"] = preview_file["duration"]
                shot["preview_file_status"] = preview_file["status"]
                shot["preview_file_annotations"] = preview_file["annotations"]
                shot["preview_file_task_id"] = preview_file["task_id"]
                shot["preview_file_previews"] = preview_file.get(
                    "previews", []
                )
            else:
                del shot["preview_file_id"]
        except Exception as e:
            print(e)
    return playlist_dict


def _add_build_job_infos_to_playlist_dict(playlist, playlist_dict):
    playlist_dict["build_jobs"] = []
    for build_job in reversed(playlist.build_jobs):
        playlist_dict["build_jobs"].append(build_job.present())
    return playlist_dict


def set_preview_files_for_entities(playlist_dict):
    """
    Retrieve all preview files related to entities listed in given playlist.
    Add to each entity a dict with task as keys and preview list as values.
    """
    entity_ids = []
    for entity in playlist_dict["shots"]:
        if "id" not in entity:
            entity_id = entity.get("shot_id", entity.get("entity_id", None))
            if entity_id is not None:
                entity_ids.append(entity_id)
                entity["id"] = entity_id
        else:
            entity_ids.append(entity["id"])
    previews = {}
    preview_file_map = {}

    preview_files = (
        PreviewFile.query.join(Task)
        .join(TaskType)
        .filter(Task.entity_id.in_(entity_ids))
        .order_by(TaskType.priority.desc())
        .order_by(TaskType.name)
        .order_by(PreviewFile.revision.desc())
        .order_by(PreviewFile.position)
        .order_by(PreviewFile.created_at)
        .add_column(Task.task_type_id)
        .add_column(Task.entity_id)
        .all()
    )

    for preview_file, task_type_id, entity_id in preview_files:
        entity_id = str(entity_id)
        task_type_id = str(task_type_id)
        if entity_id not in previews:
            previews[entity_id] = {}

        if task_type_id not in previews[entity_id]:
            previews[entity_id][task_type_id] = []

        task_id = str(preview_file.task_id)
        preview_file_id = str(preview_file.id)

        light_preview_file = {
            "id": preview_file_id,
            "revision": preview_file.revision,
            "extension": preview_file.extension,
            "width": preview_file.width,
            "height": preview_file.height,
            "duration": float(preview_file.duration or 0),
            "status": str(preview_file.status),
            "annotations": preview_file.annotations,
            "created_at": fields.serialize_value(preview_file.created_at),
            "task_id": task_id,
        }  # Do not add too much field to avoid building too big responses
        previews[entity_id][task_type_id].append(light_preview_file)
        preview_file_map[preview_file_id] = light_preview_file

    for entity_id in previews.keys():
        for task_type_id in previews[entity_id].keys():
            previews[entity_id][task_type_id] = mix_preview_file_revisions(
                previews[entity_id][task_type_id]
            )

    for entity in playlist_dict["shots"]:
        if str(entity["id"]) in previews:
            entity["preview_files"] = previews[str(entity["id"])]
        else:
            entity["preview_files"] = []

    return (fields.serialize_value(playlist_dict), preview_file_map)


def get_preview_files_for_entity(entity_id):
    """
    Get all preview files available for given shot.
    """
    previews = {}
    query = (
        Task.query.filter_by(entity_id=entity_id)
        .add_columns(
            PreviewFile.id,
            PreviewFile.revision,
            PreviewFile.position,
            PreviewFile.original_name,
            PreviewFile.extension,
            PreviewFile.width,
            PreviewFile.height,
            PreviewFile.duration,
            PreviewFile.status,
            PreviewFile.annotations,
            PreviewFile.created_at,
            PreviewFile.task_id,
        )
        .join(PreviewFile, Task.id == PreviewFile.task_id)
        .join(TaskType)
        .order_by(TaskType.priority.desc())
        .order_by(TaskType.name)
        .order_by(PreviewFile.revision.desc())
        .order_by(PreviewFile.created_at)
    )

    task_previews = {}
    for (
        task,
        preview_file_id,
        preview_file_revision,
        preview_file_position,
        preview_file_original_name,
        preview_file_extension,
        preview_file_width,
        preview_file_height,
        preview_file_duration,
        preview_file_status,
        preview_file_annotations,
        preview_file_created_at,
        preview_file_task_id,
    ) in query.all():
        task_id = str(task.id)
        if task_id not in task_previews:
            task_previews[task_id] = []
        task_previews[task_id].append(
            fields.serialize_dict(
                {
                    "id": preview_file_id,
                    "revision": preview_file_revision,
                    "position": preview_file_position,
                    "original_name": preview_file_original_name,
                    "extension": preview_file_extension,
                    "width": preview_file_width,
                    "height": preview_file_height,
                    "duration": float(preview_file_duration or 0),
                    "status": preview_file_status,
                    "annotations": preview_file_annotations,
                    "created_at": preview_file_created_at,
                    "task_id": task_id,
                    "task_type_id": str(task.task_type_id),
                }
            )
        )

    for task_id in task_previews.keys():
        preview_files = task_previews[task_id]
        task_type_id = task_previews[task_id][0]["task_type_id"]

        if len(preview_files) > 0:
            preview_files = mix_preview_file_revisions(preview_files)
            previews[task_type_id] = [
                {
                    "id": preview_file["id"],
                    "revision": preview_file["revision"],
                    "original_name": preview_file["original_name"],
                    "extension": preview_file["extension"],
                    "width": preview_file["width"],
                    "height": preview_file["height"],
                    "duration": preview_file["duration"],
                    "status": preview_file["status"],
                    "annotations": preview_file["annotations"],
                    "previews": preview_file["previews"],
                    "created_at": preview_file["created_at"],
                    "task_id": preview_file["task_id"],
                }
                for preview_file in preview_files
            ]  # Do not add too much field to avoid building too big responses
    return previews


def mix_preview_file_revisions(preview_files):
    """
    The goal here is to group preview files with same revision in a single
    preview file, which encapsulates other preview_files.
    """
    revision_map = {}
    result = []
    for preview_file in preview_files:
        revision = preview_file["revision"]
        if revision not in revision_map:
            preview_file["previews"] = []
            revision_map[revision] = preview_file
            result.append(preview_file)
        else:
            parent_preview_file = revision_map[revision]
            parent_preview_file["previews"].append(preview_file)
    return result


def get_playlist_raw(playlist_id):
    """
    Return given playlist as active record.
    """
    return base_service.get_instance(
        Playlist, playlist_id, PlaylistNotFoundException
    )


def get_playlist(playlist_id):
    """
    Return given playlist as a dict.
    """
    return get_playlist_raw(playlist_id).serialize()


def playlist_previews(shots, only_movies=False):
    """
    Retrieve all preview id and extension for the given shots.
    """
    preview_files = []
    for entity in shots:
        if (
            "preview_file_id" in entity
            and entity["preview_file_id"] is not None
            and len(entity["preview_file_id"]) > 0
        ):
            preview_file = files_service.get_preview_file(
                entity["preview_file_id"]
            )
            if preview_file is not None and (
                (only_movies and preview_file["extension"] == "mp4")
                or not only_movies
            ):
                preview_files.append(preview_file)

    return [
        {"id": x["id"], "extension": x["extension"]} for x in preview_files
    ]


def retrieve_playlist_tmp_files(preview_files, full=False):
    """
    Retrieve all files for a given playlist into the temporary folder.
    """
    file_paths = []
    for preview_file in preview_files:
        if full:
            preview_file = files_service.get_preview_file(preview_file["id"])
            sub_preview_files = (
                preview_files_service.get_preview_files_for_revision(
                    preview_file["task_id"], preview_file["revision"]
                )
            )
            for preview_file in sub_preview_files:
                tmp_file_path, file_name = retrieve_playlist_tmp_file(
                    preview_file
                )
                file_paths.append((tmp_file_path, file_name))
        else:
            tmp_file_path, file_name = retrieve_playlist_tmp_file(preview_file)
            file_paths.append((tmp_file_path, file_name))
    return file_paths


def retrieve_playlist_tmp_file(preview_file):
    if preview_file["extension"] == "mp4":
        get_path_func = file_store.get_local_movie_path
        open_func = file_store.open_movie
        exists_func = file_store.exists_movie
        prefix = "previews"
    elif preview_file["extension"] == "png":
        get_path_func = file_store.get_local_picture_path
        open_func = file_store.open_picture
        exists_func = file_store.exists_picture
        prefix = "original"
    else:
        get_path_func = file_store.get_local_file_path
        open_func = file_store.open_file
        exists_func = file_store.exists_file
        prefix = "previews"

    if config.FS_BACKEND == "local":
        file_path = get_path_func(prefix, preview_file["id"])
    else:
        file_path = os.path.join(
            config.TMP_DIR,
            "cache-previews-%s.%s"
            % (preview_file["id"], preview_file["extension"]),
        )
        if not os.path.exists(file_path) or os.path.getsize(file_path) == 0:
            if exists_func(prefix, preview_file["id"]):
                with open(file_path, "wb") as tmp_file:
                    try:
                        for chunk in open_func(prefix, preview_file["id"]):
                            tmp_file.write(chunk)
                    except FileNotFound:
                        pass
    file_name = names_service.get_preview_file_name(preview_file["id"])
    tmp_file_path = os.path.join(config.TMP_DIR, file_name)
    copyfile(file_path, tmp_file_path)
    return tmp_file_path, file_name


def build_playlist_zip_file(playlist):
    """
    Build a zip for all files for a given playlist into the temporary folder.
    """
    previews = playlist_previews(playlist["shots"])
    tmp_file_paths = retrieve_playlist_tmp_files(previews, full=True)

    zip_file_path = get_playlist_zip_file_path(playlist)
    if os.path.exists(zip_file_path):
        os.remove(zip_file_path)
    with ZipFile(zip_file_path, "w") as zip:
        for file_path, file_name in tmp_file_paths:
            zip.write(file_path, file_name)
    return zip_file_path


def build_playlist_movie_file(playlist, job, shots, params, full, remote):
    """
    Build a movie for all files for a given playlist into the temporary folder.
    """
    success = False
    from zou.app import app

    with app.app_context():
        try:
            previews = playlist_previews(shots, only_movies=True)
            movie_file_path = get_playlist_movie_file_path(job)
            tmp_file_paths = retrieve_playlist_tmp_files(previews)

            if tmp_file_paths:
                if not remote:
                    success = False
                    if not full:
                        success = _run_concatenation(
                            playlist,
                            job,
                            tmp_file_paths,
                            movie_file_path,
                            params,
                            movie.concat_demuxer,
                        )

                    # Try again using concat filter
                    if not success:
                        success = _run_concatenation(
                            playlist,
                            job,
                            tmp_file_paths,
                            movie_file_path,
                            params,
                            movie.concat_filter,
                        )
                else:
                    try:
                        _run_remote_job_build_playlist(
                            app, job, previews, params, movie_file_path, full
                        )
                        success = True
                    except Exception as exc:
                        app.logger.error(exc)
                        success = False

        except Exception as exc:
            app.logger.error(exc)
            success = False

        # exception will be logged by rq
        finally:
            job = end_build_job(playlist, job, success)

    if not success:
        raise Exception("Failure while building playlist %r" % playlist["id"])

    return job


def _run_concatenation(
    playlist, job, tmp_file_paths, movie_file_path, params, mode
):
    success = False
    try:
        result = movie.build_playlist_movie(
            mode, tmp_file_paths, movie_file_path, **params._asdict()
        )
        if result["success"] and os.path.exists(movie_file_path):
            file_store.add_movie("playlists", job["id"], movie_file_path)
            success = True
        if result.get("message"):
            from zou.app import app

            with app.app_context():
                app.logger.error(result["message"])
    except Exception:
        from zou.app import app

        with app.app_context():
            app.logger.error(
                "Unable to build playlist %r using %s",
                (playlist["id"], mode.__qualname__),
                exc_info=1,
            )
    return success


def _run_remote_job_build_playlist(
    app, job, previews, params, movie_file_path, full
):
    preview_ids = [
        preview["id"] for preview in previews if preview["extension"] == "mp4"
    ]
    input_bytes = zlib.compress(json.dumps(preview_ids))
    input_string = base64.b64encode(input_bytes).decode("ascii")
    params = {
        "version": "1",
        "output_filename": Path(movie_file_path).name,
        "output_key": file_store.make_key("playlists", job["id"]),
        "input": input_string,
        "width": params.width,
        "height": params.height,
        "fps": params.fps,
        "full": str(full).lower(),
    }
    nomad_job = config.JOB_QUEUE_NOMAD_PLAYLIST_JOB
    remote_job.run_job(app, config, nomad_job, params)

    with open(movie_file_path, "wb") as movie_file:
        for chunk in file_store.open_movie("playlists", job["id"]):
            movie_file.write(chunk)

    return movie_file_path


def start_build_job(playlist):
    """
    clients that a new job is running.
    Register in database that a new build is running. Emits an event to notify
    """
    job = BuildJob.create(
        status="running", job_type="movie", playlist_id=playlist["id"]
    )
    events.emit(
        "build-job:new",
        {
            "build_job_id": str(job.id),
            "playlist_id": playlist["id"],
            "created_at": fields.serialize_value(job.created_at),
        },
        project_id=playlist["project_id"],
    )
    return job.serialize()


def end_build_job(playlist, job, success):
    """
    Register in database that a build is finished. Emits an event to notify
    clients that the build is done.
    """
    if success:
        status = "succeeded"
    else:
        status = "failed"

    build_job = BuildJob.get(job["id"])
    if build_job is not None:
        build_job.end(status=status)
    events.emit(
        "build-job:update",
        {
            "build_job_id": job["id"],
            "playlist_id": playlist["id"],
            "status": status,
        },
        project_id=playlist["project_id"],
    )
    if build_job is not None:
        return build_job.serialize()
    else:
        return {}


def build_playlist_job(playlist, job, shots, params, email, full, remote):
    """
    Build playlist file (concatenate all movie previews). This function is
    aimed at being runned as a job in a job queue.
    """
    build_playlist_movie_file(playlist, job, shots, params, full, remote)

    # Just in case, since rq jobs which encounter an error raise an
    # exception in order to be flagged as failed.
    if job["status"] == "succeeded":
        person = persons_service.get_person_by_email_raw(email)
        organisation = persons_service.get_organisation()
        playlist_url = "%s://%s/api/data/playlists/%s/jobs/%s/build/mp4" % (
            config.DOMAIN_PROTOCOL,
            config.DOMAIN_NAME,
            playlist["id"],
            job["id"],
        )
        html = f"""<p>Hello {person.first_name},</p>
<p>Your playlist {playlist["name"]} is available at:
<a href="{playlist_url}">{playlist_url}</a>
</p>
<p>
Thank you and see you soon on Kitsu,
</p>
<p>
{organisation["name"]} Team
</p>
"""

        subject = f"{organisation['name']} - Kitsu: playlist download"
        emails.send_email(subject, html, email)


def get_playlist_file_name(playlist):
    """
    Build file name for the movie file matching given playlist.
    """
    project = projects_service.get_project(playlist["project_id"])
    download_name = "%s_%s" % (
        slugify(project["name"]),
        slugify(playlist["name"]),
    )
    return slugify(download_name)


def get_playlist_movie_file_path(build_job):
    """
    Build file path for the movie file matching given playlist.
    """
    movie_file_name = "cache-playlists-%s.mp4" % build_job["id"]
    return os.path.join(config.TMP_DIR, movie_file_name)


def get_playlist_zip_file_path(playlist):
    """
    Build file path for the archive file matching given playlist.
    """
    zip_file_name = "%s.zip" % playlist["id"]
    return os.path.join(config.TMP_DIR, zip_file_name)


def get_build_job_raw(build_job_id):
    """
    Return given build job as active record.
    """
    return base_service.get_instance(
        BuildJob, build_job_id, BuildJobNotFoundException
    )


def get_build_job(build_job_id):
    """
    Return given build job as a dict.
    """
    return get_build_job_raw(build_job_id).serialize()


def remove_playlist(playlist_id):
    """
    Remove given playlist from database (and delete related build jobs).
    """
    playlist = get_playlist_raw(playlist_id)
    playlist_dict = playlist.serialize()
    query = BuildJob.query.filter_by(playlist_id=playlist_id)
    for job in query.all():
        remove_build_job(playlist_dict, job.id)
    playlist.delete()
    events.emit(
        "playlist:delete",
        {"playlist_id": playlist_dict["id"]},
        project_id=playlist_dict["project_id"],
    )
    return playlist_dict


def remove_build_job(playlist, build_job_id):
    """
    Remove build job from database and remove related temporary file from
    hard drive.
    """
    job = BuildJob.get(build_job_id)
    movie_file_path = get_playlist_movie_file_path(job.serialize())
    if os.path.exists(movie_file_path):
        os.remove(movie_file_path)
    if config.REMOVE_FILES:
        try:
            file_store.remove_movie("playlists", build_job_id)
        except Exception:
            current_app.logger.error(
                "Playlist file can't be deleted: %s" % build_job_id
            )
    job.delete()
    events.emit(
        "build-job:delete",
        {"build_job_id": build_job_id, "playlist_id": playlist["id"]},
        project_id=playlist["project_id"],
    )
    return movie_file_path


def get_build_jobs_for_project(project_id):
    """
    Return all build_jobs for given project.
    """
    build_jobs = BuildJob.query.join(Playlist).filter(
        Playlist.project_id == project_id
    )
    return fields.serialize_list(build_jobs)


def get_playlists_for_project(project_id, page=0):
    """
    Return all playlists for given project.
    """
    query = Playlist.query.filter(Playlist.project_id == project_id)
    return query_utils.get_paginated_results(query, page, relations=True)


def generate_temp_playlist(task_ids, sort=True):
    """
    Generate the data structure of a playlist for a given task list. It doesn't
    persist anything. The goal is to build a temporary playlist used to see
    a quick preview of several shots.
    """
    entities = []
    for task_id in task_ids:
        entity = generate_playlisted_entity_from_task(task_id)
        entities.append(entity)
    if len(entities) > 0:
        if not sort:
            return entities
        try:
            if "episode_name" in entities[0]:
                return sorted(entities, key=itemgetter("episode_name", "name"))
            elif "sequence_name" in entities[0]:
                return sorted(
                    entities, key=itemgetter("sequence_name", "name")
                )
            elif "asset_type_name" in entities[0]:
                return sorted(
                    entities, key=itemgetter("asset_type_name", "name")
                )
            else:
                return entities
        except BaseException:
            return entities
    else:
        return []


def generate_playlisted_entity_from_task(task_id):
    """
    Generate the data structure of a playlisted shot for a given task. It
    doesn't persist anything.
    """
    task = tasks_service.get_task(task_id)
    task = tasks_service.get_task(task_id)
    entity = entities_service.get_entity(task["entity_id"])
    task = tasks_service.get_task(task_id)
    if shots_service.is_shot(entity):
        playlisted_entity = get_base_shot_for_playlist(entity, task_id)
    elif shots_service.is_sequence(entity):
        playlisted_entity = get_base_sequence_for_playlist(entity, task_id)
    elif shots_service.is_episode(entity):
        playlisted_entity = get_base_episode_for_playlist(entity, task_id)
    else:
        playlisted_entity = get_base_asset_for_playlist(entity, task_id)

    task_type_id = task["task_type_id"]
    preview_files = get_preview_files_for_entity(entity["id"])
    if task_type_id in preview_files and len(preview_files[task_type_id]) > 0:
        preview_file = preview_files[task_type_id][0]
        playlisted_entity.update(
            {
                "preview_file_id": preview_file["id"],
                "preview_file_extension": preview_file["extension"],
                "preview_file_width": preview_file["width"],
                "preview_file_height": preview_file["height"],
                "preview_file_duration": preview_file["duration"],
                "preview_file_revision": preview_file["revision"],
                "preview_file_status": preview_file["status"],
                "preview_file_annotations": preview_file["annotations"],
                "preview_file_previews": preview_file["previews"],
            }
        )
    playlisted_entity["preview_files"] = preview_files
    return playlisted_entity


def get_base_episode_for_playlist(entity, task_id):
    episode = shots_service.get_episode(entity["id"])
    playlisted_entity = {
        "id": episode["id"],
        "name": episode["name"],
        "preview_file_task_id": task_id,
        "sequence_id": "",
        "sequence_name": "",
        "parent_name": "",
    }
    return playlisted_entity


def get_base_sequence_for_playlist(entity, task_id):
    sequence = shots_service.get_sequence(entity["id"])
    episode = None
    try:
        episode = shots_service.get_episode(sequence["parent_id"])
    except:
        pass
    if episode is not None:
        playlisted_entity = {
            "id": sequence["id"],
            "name": sequence["name"],
            "preview_file_task_id": task_id,
            "episode_id": episode["id"],
            "episode_name": episode["name"],
            "parent_name": episode["name"],
        }
    else:
        playlisted_entity = {
            "id": sequence["id"],
            "name": sequence["name"],
            "preview_file_task_id": task_id,
            "sequence_id": "",
            "sequence_name": "",
            "parent_name": "",
        }
    return playlisted_entity


def get_base_shot_for_playlist(entity, task_id):
    shot = shots_service.get_shot(entity["id"])
    sequence = shots_service.get_sequence(shot["parent_id"])
    playlisted_entity = {
        "id": shot["id"],
        "name": shot["name"],
        "preview_file_task_id": task_id,
        "sequence_id": sequence["id"],
        "sequence_name": sequence["name"],
        "parent_name": sequence["name"],
    }
    return playlisted_entity


def get_base_asset_for_playlist(entity, task_id):
    asset = assets_service.get_asset(entity["id"])
    asset_type = assets_service.get_asset_type(asset["entity_type_id"])
    playlisted_entity = {
        "id": asset["id"],
        "name": asset["name"],
        "preview_file_task_id": task_id,
        "asset_type_id": asset_type["id"],
        "asset_type_name": asset_type["name"],
        "parent_name": asset_type["name"],
    }
    return playlisted_entity


def get_preview_files_for_task(task_id):
    """
    Return all preview file active records for given task.
    """
    preview_files = (
        PreviewFile.query.filter_by(task_id=task_id)
        .order_by(PreviewFile.revision.desc())
        .all()
    )
    return _get_playlist_preview_file_list(preview_files)


def _get_playlist_preview_file_list(preview_files):
    """
    Turn preview file active records into preview file dict that match the
    playlist data structure.
    """
    return [
        {
            "id": str(preview_file.id),
            "revision": preview_file.revision,
            "extension": preview_file.extension,
            "width": preview_file.width,
            "height": preview_file.height,
            "duration": float(preview_file.duration or 0),
            "status": str(preview_file.status),
            "annotations": preview_file.annotations,
            "created_at": fields.serialize_value(preview_file.created_at),
            "task_id": str(preview_file.task_id),
        }
        for preview_file in preview_files
    ]



from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource

from zou.app.models.entity_type import EntityType
from zou.app.utils import events
from zou.app.services import entities_service, assets_service

from zou.app.services.exception import WrongParameterException


class EntityTypesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, EntityType)

    def all_entries(self, query=None, relations=False):
        if query is None:
            query = self.model.query

        return [
            asset_type.serialize(relations=relations)
            for asset_type in query.all()
        ]

    def check_read_permissions(self, options=None):
        return True

    def emit_create_event(self, instance_dict):
        events.emit("asset-type:new", {"asset_type_id": instance_dict["id"]})

    def post_creation(self, instance):
        assets_service.clear_asset_type_cache()
        return instance.serialize(relations=True)

    def check_creation_integrity(self, data):
        entity_type = EntityType.query.filter(
            EntityType.name.ilike(data.get("name", ""))
        ).first()
        if entity_type is not None:
            raise WrongParameterException(
                "Entity type with this name already exists"
            )
        return data


class EntityTypeResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, EntityType)

    def check_read_permissions(self, instance):
        return True

    def emit_update_event(self, instance_dict):
        events.emit(
            "asset-type:update", {"asset_type_id": instance_dict["id"]}
        )

    def emit_delete_event(self, instance_dict):
        events.emit(
            "asset-type:delete", {"asset_type_id": instance_dict["id"]}
        )

    def post_update(self, instance_dict, data):
        entities_service.clear_entity_type_cache(instance_dict["id"])
        assets_service.clear_asset_type_cache()
        instance_dict["task_types"] = [
            str(task_types.id) for task_types in self.instance.task_types
        ]
        return instance_dict

    def post_delete(self, instance_dict):
        entities_service.clear_entity_type_cache(instance_dict["id"])
        assets_service.clear_asset_type_cache()
        return instance_dict



"""empty message

Revision ID: 925771029620
Revises: fb6b6f188497
Create Date: 2018-09-26 21:16:04.776712

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "925771029620"
down_revision = "fb6b6f188497"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("playlist_uc", "playlist", type_="unique")
    op.create_unique_constraint(
        "playlist_uc", "playlist", ["name", "project_id", "episode_id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("playlist_uc", "playlist", type_="unique")
    op.create_unique_constraint(
        "playlist_uc", "playlist", ["name", "project_id"]
    )
    # ### end Alembic commands ###



"""Fix schedule item

Revision ID: 10cf267d95c9
Revises: b45cb782bb9c
Create Date: 2019-08-21 17:11:00.542481

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
from sqlalchemy.dialects import postgresql
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "10cf267d95c9"
down_revision = "b45cb782bb9c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "schedule_item",
        sa.Column(
            "object_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_index(
        op.f("ix_schedule_item_object_id"),
        "schedule_item",
        ["object_id"],
        unique=False,
    )
    op.drop_index("ix_schedule_item_entity_id", table_name="schedule_item")
    op.drop_constraint(
        "schedule_item_entity_id_fkey", "schedule_item", type_="foreignkey"
    )
    op.drop_column("schedule_item", "entity_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "schedule_item",
        sa.Column(
            "entity_id", postgresql.UUID(), autoincrement=False, nullable=True
        ),
    )
    op.create_foreign_key(
        "schedule_item_entity_id_fkey",
        "schedule_item",
        "entity",
        ["entity_id"],
        ["id"],
    )
    op.create_index(
        "ix_schedule_item_entity_id",
        "schedule_item",
        ["entity_id"],
        unique=False,
    )
    op.drop_index(
        op.f("ix_schedule_item_object_id"), table_name="schedule_item"
    )
    op.drop_column("schedule_item", "object_id")
    # ### end Alembic commands ###



from zou.app.models.chat_message import ChatMessage

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class ChatMessagesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, ChatMessage)


class ChatMessageResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, ChatMessage)



from sqlalchemy.exc import StatementError

from zou.app.models.project import Project
from zou.app.models.entity import Entity
from zou.app.models.notification import Notification
from zou.app.models.subscription import Subscription
from zou.app.models.task import Task
from zou.app.models.task_type import TaskType

from zou.app.services import (
    assets_service,
    emails_service,
    projects_service,
    tasks_service,
    persons_service,
)
from zou.app.services.exception import PersonNotFoundException
from zou.app.utils import events, fields, query as query_utils

from zou.app.utils import cache


@cache.memoize_function(120)
def is_person_subscribed(person_id, task_id):
    """
    Returns true if the user subscribed to given task notifications.
    """
    subscription = Subscription.get_by(task_id=task_id, person_id=person_id)
    return subscription is not None


def create_notification(
    person_id,
    comment_id=None,
    author_id=None,
    task_id=None,
    reply_id=None,
    read=False,
    change=False,
    type="comment",
    created_at=None,
):
    """
    Create a new notification for given person and comment.
    """
    creation_date = fields.get_default_date_object(created_at)
    notification = Notification.create(
        read=read,
        change=change,
        person_id=person_id,
        author_id=author_id,
        task_id=task_id,
        comment_id=comment_id,
        reply_id=reply_id,
        type=type,
        created_at=creation_date,
    )
    return notification.serialize()


def get_notification_recipients(task, replies=[]):
    """
    Get the list of notification recipients for given task: assignees and
    every people who commented the task.
    """
    recipients = set()
    task_subscriptions = get_task_subscriptions(task)
    sequence_subscriptions = get_sequence_subscriptions(task)

    for assignee_id in task["assignees"]:
        recipients.add(assignee_id)

    for subscription in task_subscriptions:
        recipients.add(str(subscription.person_id))

    for subscription in sequence_subscriptions:
        recipients.add(str(subscription.person_id))

    for reply in replies:
        recipients.add(reply["person_id"])

    return recipients


def get_task_subscriptions(task):
    """
    Return all notification subscriptions related to given task.
    """
    return Subscription.get_all_by(task_id=task["id"])


def get_sequence_subscriptions(task):
    """
    Return all sequence subscriptions for given task. It returns something only
    if the task is related to a shot of which the sequence has a subscription.
    """
    sequence_subscriptions = []
    entity = Entity.get(task["entity_id"])
    if entity is not None and entity.parent_id is not None:
        sequence_subscriptions = Subscription.get_all_by(
            task_type_id=task["task_type_id"], entity_id=entity.parent_id
        )
    return sequence_subscriptions


def create_notifications_for_task_and_comment(task, comment, change=False):
    """
    For given task and comment, create a notification for every assignee
    to the task and to every person participating to this task.
    """
    recipient_ids = get_notification_recipients(task)
    if comment["person_id"] in recipient_ids:
        recipient_ids.remove(comment["person_id"])
    author_id = comment["person_id"]
    task = tasks_service.get_task(comment["object_id"])

    for recipient_id in recipient_ids:
        try:
            notification = create_notification(
                recipient_id,
                comment_id=comment["id"],
                author_id=author_id,
                task_id=task["id"],
                read=False,
                change=change,
                type="comment",
            )
            emails_service.send_comment_notification(
                recipient_id, author_id, comment, task
            )
            events.emit(
                "notification:new",
                {
                    "notification_id": notification["id"],
                    "person_id": recipient_id,
                },
                project_id=task["project_id"],
                persist=False,
            )
        except PersonNotFoundException:
            pass

    mentions = get_mentioned_people(task["project_id"], comment)
    for recipient_id in mentions:
        if recipient_id != comment["person_id"]:
            notification = create_notification(
                recipient_id,
                comment_id=comment["id"],
                author_id=comment["person_id"],
                task_id=comment["object_id"],
                type="mention",
            )
            emails_service.send_mention_notification(
                recipient_id, author_id, comment, task
            )
            events.emit(
                "notification:new",
                {
                    "notification_id": notification["id"],
                    "person_id": recipient_id,
                },
                project_id=task["project_id"],
                persist=False,
            )

    return recipient_ids


def get_mentioned_people(project_id, comment):
    """
    Return all people mentioned in the comment: the one listed via their name
    and the one listed via their department.
    """
    mentions = comment["mentions"]
    for department_id in comment["department_mentions"]:
        persons = projects_service.get_department_team(
            project_id, department_id
        )
        for person in persons:
            mentions.append(str(person.id))
    return mentions


def create_notifications_for_task_and_reply(task, comment, reply):
    """
    For given task, comment and reply, create a notification for every assignee
    to the task and to every person participating to this task and comment.
    """
    recipient_ids = get_notification_recipients(task, comment["replies"])
    if reply["person_id"] in recipient_ids:
        recipient_ids.remove(reply["person_id"])
    author_id = reply["person_id"]
    if author_id != comment["person_id"]:
        recipient_ids.add(comment["person_id"])
    task = tasks_service.get_task(comment["object_id"])
    for recipient_id in recipient_ids:
        try:
            notification = create_notification(
                recipient_id,
                comment_id=comment["id"],
                author_id=author_id,
                task_id=task["id"],
                reply_id=reply["id"],
                read=False,
                type="reply",
                created_at=reply["created_at"],
            )
            emails_service.send_reply_notification(
                recipient_id, author_id, comment, task, reply
            )
            events.emit(
                "notification:new",
                {
                    "notification_id": notification["id"],
                    "person_id": recipient_id,
                },
                project_id=task["project_id"],
                persist=False,
            )
        except PersonNotFoundException:
            pass

    mentions = get_mentioned_people(task["project_id"], reply)
    for recipient_id in mentions:
        if recipient_id != reply["person_id"]:
            notification = create_notification(
                recipient_id,
                comment_id=comment["id"],
                author_id=reply["person_id"],
                task_id=task["id"],
                reply_id=reply["id"],
                read=False,
                type="reply-mention",
            )
            emails_service.send_mention_notification(
                recipient_id, author_id, comment, task
            )
            events.emit(
                "notification:new",
                {
                    "notification_id": notification["id"],
                    "person_id": recipient_id,
                },
                project_id=task["project_id"],
                persist=False,
            )

    return recipient_ids


def reset_notifications_for_mentions(comment):
    """
    For given task and comment, delete all mention notifications related
    to the comment and recreate notifications for the mentions listed in the
    comment.
    """
    Notification.delete_all_by(type="mention", comment_id=comment["id"])
    Notification.delete_all_by(type="reply", comment_id=comment["id"])
    notifications = []
    task = tasks_service.get_task(comment["object_id"])
    author_id = comment["person_id"]
    mentions = get_mentioned_people(task["project_id"], comment)
    for recipient_id in mentions:
        notification = create_notification(
            recipient_id,
            comment_id=comment["id"],
            author_id=author_id,
            task_id=comment["object_id"],
            type="mention",
            created_at=comment["created_at"],
        )
        emails_service.send_mention_notification(
            recipient_id, author_id, comment, task
        )
        notifications.append(notification)
        events.emit(
            "notification:new",
            {"notification_id": notification["id"], "person_id": recipient_id},
            project_id=task["project_id"],
            persist=False,
        )
    return notifications


def create_assignation_notification(task_id, person_id, author_id=None):
    """
    Create a notification following a task assignation.
    """
    task = tasks_service.get_task_raw(task_id)
    if author_id is None:
        author_id = task.assigner_id

    if str(author_id) != person_id:
        notification = create_notification(
            person_id, author_id=author_id, task_id=task_id, type="assignation"
        )
        emails_service.send_assignation_notification(
            person_id, author_id, task.serialize()
        )
        events.emit(
            "notification:new",
            {"notification_id": notification["id"], "person_id": person_id},
            project_id=str(task.project_id),
            persist=False,
        )
        return notification
    else:
        return None


def get_task_subscription_raw(person_id, task_id):
    """
    Return subscription matching given person and task.
    """
    try:
        subscription = Subscription.get_by(
            person_id=person_id, task_id=task_id
        )
        return subscription
    except StatementError:
        return None


def has_task_subscription(person_id, task_id):
    """
    Return true if a subscription exists for this person and this task.
    """
    subscription = get_task_subscription_raw(person_id, task_id)
    return subscription is not None


def subscribe_to_task(person_id, task_id):
    """
    Add a subscription entry for given person and task.
    """
    subscription = get_task_subscription_raw(person_id, task_id)
    if subscription is None:
        subscription = Subscription.create(
            person_id=person_id, task_id=task_id
        )
    cache.cache.delete_memoized(is_person_subscribed, person_id, task_id)
    return subscription.serialize()


def unsubscribe_from_task(person_id, task_id):
    """
    Remove subscription entry for given person and task.
    """
    subscription = get_task_subscription_raw(person_id, task_id)
    if subscription is not None:
        subscription.delete()
        cache.cache.delete_memoized(is_person_subscribed, person_id, task_id)
        return subscription.serialize()
    else:
        return {}


def get_sequence_subscription_raw(person_id, sequence_id, task_type_id):
    """
    Return subscription matching given person, sequence and task type.
    """
    try:
        subscription = Subscription.get_by(
            person_id=person_id,
            entity_id=sequence_id,
            task_type_id=task_type_id,
        )
        return subscription
    except StatementError:
        return None


def has_sequence_subscription(person_id, sequence_id, task_type_id):
    """
    Return true if a subscription exists for this person, sequence and task
    type.
    """
    subscription = get_sequence_subscription_raw(
        person_id, sequence_id, task_type_id
    )
    return subscription is not None


def subscribe_to_sequence(person_id, sequence_id, task_type_id):
    """
    Add a subscription entry for given person, sequence and task type.
    """
    subscription = get_sequence_subscription_raw(
        person_id, sequence_id, task_type_id
    )
    if subscription is None:
        subscription = Subscription.create(
            person_id=person_id,
            entity_id=sequence_id,
            task_type_id=task_type_id,
        )
    return subscription.serialize()


def unsubscribe_from_sequence(person_id, sequence_id, task_type_id):
    """
    Remove subscription entry for given person, sequence and task type.
    """
    subscription = get_sequence_subscription_raw(
        person_id, sequence_id, task_type_id
    )
    if subscription is not None:
        subscription.delete()
        return subscription.serialize()
    else:
        return {}


def get_all_sequence_subscriptions(person_id, project_id, task_type_id):
    """
    Return list of sequence ids for which given person has subscribed for
    given project and task type.
    """
    subscriptions = (
        Subscription.query.join(Entity)
        .join(Project)
        .filter(Project.id == project_id)
        .filter(TaskType.id == task_type_id)
        .all()
    )

    return fields.serialize_value(
        [subscription.entity_id for subscription in subscriptions]
    )


def delete_notifications_for_comment(comment_id):
    notifications = Notification.get_all_by(comment_id=comment_id)
    for notification in notifications:
        notification.delete()
    return fields.serialize_list(notifications)


def get_last_notifications(notification_type=None):
    """
    Return last notification created. This function is used mainly for testing
    purpose.
    """
    query = Notification.query
    if notification_type is not None:
        query = query.filter_by(type=notification_type)
    return fields.serialize_value(query.limit(100).all())


def get_subscriptions_for_project(project_id):
    """
    Return all subscriptions for given project.
    """
    subscriptions = Subscription.query.join(Task).filter(
        Task.project_id == project_id
    )
    return fields.serialize_list(subscriptions)


def get_notifications_for_project(project_id, page=0):
    """
    Return all notifications for given project.
    """
    query = (
        Notification.query.join(Task)
        .filter(Task.project_id == project_id)
        .order_by(Notification.updated_at.desc())
    )
    return query_utils.get_paginated_results(query, page)


def get_subscriptions_for_user(project_id, entity_type_id=None):
    subscription_map = {}
    if project_id is not None:
        user_id = persons_service.get_current_user()["id"]
        if entity_type_id is not None:
            subscriptions = (
                Subscription.query.join(Task)
                .join(Entity, Task.entity_id == Entity.id)
                .filter(Subscription.person_id == user_id)
                .filter(Entity.entity_type_id == entity_type_id)
                .filter(Task.project_id == project_id)
            ).all()
        else:
            subscriptions = (
                Subscription.query.join(Task)
                .join(Entity, Task.entity_id == Entity.id)
                .filter(Subscription.person_id == user_id)
                .filter(Task.project_id == project_id)
                .filter(assets_service.build_asset_type_filter())
            ).all()
        for subscription in subscriptions:
            subscription_map[str(subscription.task_id)] = True
    return subscription_map



"""add task type field to playlists

Revision ID: 6c597e842afa
Revises: 6d7fa5a8e9a5
Create Date: 2021-05-10 00:12:27.927441

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "6c597e842afa"
down_revision = "6d7fa5a8e9a5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "playlist",
        sa.Column(
            "task_type_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_index(
        op.f("ix_playlist_task_type_id"),
        "playlist",
        ["task_type_id"],
        unique=False,
    )
    op.create_foreign_key(
        None, "playlist", "task_type", ["task_type_id"], ["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "playlist", type_="foreignkey")
    op.drop_index(op.f("ix_playlist_task_type_id"), table_name="playlist")
    op.drop_column("playlist", "task_type_id")
    # ### end Alembic commands ###



"""empty message

Revision ID: 443d1e78a932
Revises: 59e149a772cf
Create Date: 2018-08-07 21:36:03.384838

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "443d1e78a932"
down_revision = "c726b98be194"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_type", sa.Column("allow_timelog", sa.Boolean(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_type", "allow_timelog")
    # ### end Alembic commands ###



import os
import json
import sys

from pathlib import Path

from flask_fs import Storage


def get_config_from_payload():
    if len(sys.argv) < 2:
        print("Required parameter is missing.", file=sys.stderr)
        sys.exit(1)

    payload_file = Path(sys.argv[1])
    if not payload_file.exists():
        print("Payload file %r doesn't exist" % sys.argv[1], file=sys.stderr)
        sys.exit(1)

    with payload_file.open() as data:
        config = json.load(data)

    return config


def check_config_version(config):
    try:
        version = int(config["version"])
    except (ValueError, KeyError):
        version = None
    if version is None or version > 1:
        print(
            "Input parameters: unsupported format (version: %r)" % version,
            file=sys.stderr,
        )
        sys.exit(1)
    return version


class FakeApp(object):
    def __init__(self, config):
        self.config = config


def get_storage(config, bucket):
    if config["FS_BACKEND"] not in ["s3", "swift"]:
        print("Unknown object storage backend (%r)" % config["FS_BACKEND"])
        sys.exit(1)

    storage = Storage(
        "%s%s" % (config.get("FS_BUCKET_PREFIX", ""), bucket),
        overwrite=True,
    )
    app = FakeApp(config)
    storage.configure(app)

    return storage


def make_key(prefix, id):
    return f"{prefix}-{id}"


def get_file_from_storage(storage, output_file_path, filename):
    if (
        not os.path.isfile(output_file_path)
        or os.path.getsize(output_file_path) == 0
    ):
        with open(output_file_path, "wb") as output_file:
            for chunk in storage.read_chunks(filename):
                output_file.write(chunk)
    return output_file_path


def put_file_to_storage(storage, input_file_path, filename):
    with open(input_file_path, "rb") as input_file:
        storage.write(filename, input_file)
    return input_file_path



from zou.app.blueprints.source.csv.base import (
    BaseCsvProjectImportResource,
    RowException,
)
from zou.app.models.project import ProjectTaskTypeLink
from zou.app.models.task_type import TaskType

from zou.app.services import (
    edits_service,
    projects_service,
    shots_service,
    persons_service,
)
from zou.app.models.entity import Entity
from zou.app.services.tasks_service import (
    create_task,
    create_tasks,
    get_tasks_for_edit,
    get_task_statuses,
    get_task_type,
)
from zou.app.services.comments_service import create_comment
from zou.app.services.exception import WrongParameterException
from zou.app.utils import events, string


class EditsCsvImportResource(BaseCsvProjectImportResource):
    def post(self, project_id):
        """
        Import project edits.
        ---
        tags:
          - Import
        consumes:
          - multipart/form-data
        parameters:
          - in: formData
            name: file
            type: file
            required: true
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Edits imported
            400:
                description: Format error
        """
        return super().post(project_id)

    def prepare_import(self, project_id):
        self.episodes = {}
        self.entity_types = {}
        self.descriptor_fields = self.get_descriptor_field_map(
            project_id, "Edit"
        )
        project = projects_service.get_project(project_id)
        self.is_tv_show = projects_service.is_tv_show(project)
        if self.is_tv_show:
            episodes = shots_service.get_episodes_for_project(project_id)
            self.episodes = {
                episode["name"]: episode["id"] for episode in episodes
            }
        self.created_edits = []
        self.task_types_in_project_for_edits = (
            TaskType.query.join(ProjectTaskTypeLink)
            .filter(ProjectTaskTypeLink.project_id == project_id)
            .filter(TaskType.for_entity == "Edit")
        )
        self.task_statuses = {
            status["id"]: [status[n].lower() for n in ("name", "short_name")]
            for status in get_task_statuses()
        }
        self.current_user_id = persons_service.get_current_user()["id"]

    def get_tasks_update(self, row):
        tasks_update = []
        for task_type in self.task_types_in_project_for_edits:
            task_status_name = row.get(task_type.name, None)
            task_status_id = None
            if task_status_name not in [None, ""]:
                for status_id, status_names in self.task_statuses.items():
                    if task_status_name.lower() in status_names:
                        task_status_id = status_id
                        break
                if task_status_id is None:
                    raise RowException(
                        "Task status not found for %s" % task_status_name
                    )

            task_comment_text = row.get("%s comment" % task_type.name, None)

            if task_status_id is not None or task_comment_text not in [
                None,
                "",
            ]:
                tasks_update.append(
                    {
                        "task_type_id": str(task_type.id),
                        "task_status_id": task_status_id,
                        "comment": task_comment_text,
                    }
                )

        return tasks_update

    def create_and_update_tasks(
        self, tasks_update, entity, edit_creation=False
    ):
        if tasks_update:
            if edit_creation:
                tasks_map = {
                    str(task_type.id): create_task(
                        task_type.serialize(), entity.serialize()
                    )
                    for task_type in self.task_types_in_project_for_edits
                }
            else:
                tasks_map = {
                    task["task_type_id"]: task
                    for task in get_tasks_for_edit(str(entity.id))
                }

            for task_update in tasks_update:
                if task_update["task_type_id"] not in tasks_map:
                    tasks_map[task_update["task_type_id"]] = create_task(
                        get_task_type(task_update["task_type_id"]),
                        entity.serialize(),
                    )
                task = tasks_map[task_update["task_type_id"]]
                if (
                    task_update["comment"] is not None
                    or task_update["task_status_id"] != task["task_status_id"]
                ):
                    try:
                        create_comment(
                            self.current_user_id,
                            task["id"],
                            task_update["task_status_id"]
                            or task["task_status_id"],
                            task_update["comment"] or "",
                            [],
                            {},
                            "",
                        )
                    except WrongParameterException:
                        pass
        elif edit_creation:
            self.created_edits.append(entity.serialize())

    def import_row(self, row, project_id):
        edit_name = row["Name"]
        episode_name = row.get("Episode", None)
        episode_id = None

        if self.is_tv_show:
            if episode_name is not None and episode_name not in list(
                self.episodes.keys()
            ):
                self.episodes[episode_name] = shots_service.create_episode(
                    project_id, episode_name, created_by=self.current_user_id
                )["id"]
            episode_id = self.episodes.get(episode_name, None)
        elif episode_name is not None:
            raise RowException(
                "An episode column is present for a production that isn't a TV Show"
            )

        edit_type_id = edits_service.get_edit_type()["id"]

        edit_values = {
            "name": edit_name,
            "project_id": project_id,
            "entity_type_id": edit_type_id,
            "parent_id": episode_id,
        }

        entity = Entity.get_by(**edit_values)

        edit_new_values = {}

        description = row.get("Description", None)
        if description is not None:
            edit_new_values["description"] = description

        if entity is None or not entity.data:
            edit_new_values["data"] = {}
        else:
            edit_new_values["data"] = entity.data.copy()

        for name, descriptor in self.descriptor_fields.items():
            if name in row:
                if descriptor["data_type"] == "boolean":
                    edit_new_values["data"][descriptor["field_name"]] = (
                        "true" if string.strtobool(row[name]) else "false"
                    )
                else:
                    edit_new_values["data"][descriptor["field_name"]] = row[
                        name
                    ]

        tasks_update = self.get_tasks_update(row)

        if entity is None:
            entity = Entity.create(
                **{**edit_values, **edit_new_values},
                created_by=self.current_user_id
            )
            events.emit(
                "edit:new",
                {"edit_id": str(entity.id), "episode_id": episode_id},
                project_id=project_id,
            )

            self.create_and_update_tasks(
                tasks_update, entity, edit_creation=True
            )

        elif self.is_update:
            entity.update(edit_new_values)
            events.emit(
                "edit:update",
                {"edit_id": str(entity.id), "episode_id": episode_id},
                project_id=project_id,
            )

            self.create_and_update_tasks(
                tasks_update, entity, edit_creation=False
            )

        return entity.serialize()

    def run_import(self, file_path, project_id):
        entities = super().run_import(file_path, project_id)
        for task_type in self.task_types_in_project_for_edits:
            create_tasks(task_type.serialize(), self.created_edits)
        return entities



from zou.app import config


def get_redis_url(db_index):
    redis_host = config.KEY_VALUE_STORE["host"]
    redis_port = config.KEY_VALUE_STORE["port"]
    if config.KEY_VALUE_STORE["password"]:
        redis_password = f":{config.KEY_VALUE_STORE['password']}@"
    else:
        redis_password = ""
    return f"redis://{redis_password}{redis_host}:{redis_port}/{db_index}"



import pyotp
import random
import string
import flask_bcrypt

from datetime import timedelta

from flask import request, session, current_app
from babel.dates import format_datetime

from ldap3 import Server, Connection, ALL, NTLM, SIMPLE
from ldap3.core.exceptions import (
    LDAPSocketOpenError,
    LDAPInvalidCredentialsResult,
)

from zou.app.services import persons_service
from zou.app.models.person import Person
from zou.app.services.exception import (
    EmailOTPAlreadyEnabledException,
    EmailOTPNotEnabledException,
    FIDONoPreregistrationException,
    FIDOServerException,
    MissingOTPException,
    NoAuthStrategyConfigured,
    PersonNotFoundException,
    TooMuchLoginFailedAttemps,
    TOTPAlreadyEnabledException,
    TOTPNotEnabledException,
    TwoFactorAuthenticationNotEnabledException,
    UnactiveUserException,
    UserCantConnectDueToNoFallback,
    WrongOTPException,
    WrongPasswordException,
    WrongUserException,
)
from zou.app.stores import auth_tokens_store
from zou.app.utils import date_helpers, emails

from fido2.webauthn import (
    PublicKeyCredentialUserEntity,
)
from sqlalchemy.orm.attributes import flag_modified

from fido2.utils import bytes2int, int2bytes
from fido2.webauthn import AttestedCredentialData


def check_auth(
    app,
    email,
    password,
    totp=None,
    email_otp=None,
    fido_authentication_response=None,
    recovery_code=None,
    no_otp=False,
):
    """
    Depending on configured strategy, it checks if given email and password
    mach an active user in the database. It raises exceptions adapted to
    encountered error (no auth strategy configured, wrong email, wrong passwor
    or unactive user).
    App is needed as parameter to give access to configuration while avoiding
    cyclic imports.
    """
    if not email:
        raise WrongUserException()
    try:
        person = persons_service.get_person_by_email_desktop_login(email)
    except PersonNotFoundException:
        raise WrongUserException()

    if not person.get("active", False):
        raise UnactiveUserException()

    login_failed_attemps = check_login_failed_attemps(person)

    strategy = app.config["AUTH_STRATEGY"]
    try:
        if strategy == "auth_local_classic":
            local_auth_strategy(person, password, app)
        elif strategy == "auth_local_no_password":
            no_password_auth_strategy(person, password, app)
        elif strategy == "auth_remote_ldap":
            ldap_auth_strategy(person, password, app)
        else:
            raise NoAuthStrategyConfigured()
    except WrongPasswordException:
        update_login_failed_attemps(
            person["id"],
            login_failed_attemps + 1,
            date_helpers.get_utc_now_datetime(),
        )
        raise WrongPasswordException()

    if not no_otp and person_two_factor_authentication_enabled(person):
        if not check_two_factor_authentication(
            person,
            totp,
            email_otp,
            fido_authentication_response,
            recovery_code,
        ):
            update_login_failed_attemps(
                person["id"],
                login_failed_attemps + 1,
                date_helpers.get_utc_now_datetime(),
            )
            raise WrongOTPException()

    if login_failed_attemps > 0:
        update_login_failed_attemps(person["id"], 0)

    if "password" in person:
        del person["password"]

    if "totp_secret" in person:
        del person["totp_secret"]

    if "email_otp_secret" in person:
        del person["email_otp_secret"]

    if "otp_recovery_codes" in person:
        del person["otp_recovery_codes"]

    if "fido_credentials" in person:
        del person["fido_credentials"]

    return person


def no_password_auth_strategy(person, password, app):
    """
    No password auth strategy
    """
    return person


def local_auth_strategy(person, password, app=None):
    """
    Local strategy just checks that person and passwords are correct the
    traditional way (email is in database and related password hash corresponds
    to given password).
    Password hash comparison is based on BCrypt.
    """
    try:
        password_hash = person["password"] or ""
        if password_hash and flask_bcrypt.check_password_hash(
            password_hash, password
        ):
            return person
        else:
            raise WrongPasswordException()
    except ValueError:
        raise WrongPasswordException()


def ldap_auth_strategy(person, password, app):
    """
    Connect to an active directory server to know if given user can be
    authenticated.
    When person is not from ldap, it can try to connect with default auth
    strategy.
    (only if fallback is activated (via LDAP_FALLBACK flag) in configuration)
    """
    if person["is_generated_from_ldap"]:
        try:
            SSL = app.config["LDAP_SSL"]
            if app.config["LDAP_IS_AD_SIMPLE"]:
                user = f"cn={person['desktop_login']},{app.config['LDAP_BASE_DN']}"
                authentication = SIMPLE
            elif app.config["LDAP_IS_AD"]:
                user = (
                    f"{app.config['LDAP_DOMAIN']}\\{person['desktop_login']}"
                )
                authentication = NTLM
            else:
                user = f"uid={person['desktop_login']},{app.config['LDAP_BASE_DN']}"
                authentication = SIMPLE

            ldap_server = (
                f"{app.config['LDAP_HOST']}:{app.config['LDAP_PORT']}"
            )
            server = Server(ldap_server, get_info=ALL, use_ssl=SSL)
            conn = Connection(
                server,
                user=user,
                password=password,
                authentication=authentication,
                raise_exceptions=True,
            )
            conn.bind()
            return person

        except LDAPSocketOpenError:
            app.logger.error(
                "Cannot connect to LDAP/Active directory server %s "
                % (ldap_server)
            )
            raise LDAPSocketOpenError()

        except LDAPInvalidCredentialsResult:
            app.logger.error(
                "LDAP cannot authenticate user: %s" % person["email"]
            )
            raise WrongPasswordException()

    elif app.config["LDAP_FALLBACK"]:
        return local_auth_strategy(person, password, app)
    else:
        raise UserCantConnectDueToNoFallback()


def update_login_failed_attemps(
    person_id, login_failed_attemps, last_login_failed=None
):
    """
    Update login failed attemps for a person_id.
    """
    person = Person.get(person_id)
    person.login_failed_attemps = login_failed_attemps
    if last_login_failed is not None:
        person.last_login_failed = last_login_failed
    person.commit()
    persons_service.clear_person_cache()
    return person.serialize()


def check_two_factor_authentication(
    person,
    totp=None,
    email_otp=None,
    fido_authentication_response=None,
    recovery_code=None,
):
    """
    Check multifactor authentication for a person.
    """
    if person["totp_enabled"] and totp is not None:
        return check_totp(person, totp)
    elif person["email_otp_enabled"] and email_otp is not None:
        return check_email_otp(person, email_otp)
    elif person["fido_enabled"] and fido_authentication_response is not None:
        return check_fido(person, fido_authentication_response)
    elif recovery_code is not None:
        return check_recovery_code(person, recovery_code)
    else:
        raise MissingOTPException(
            person["preferred_two_factor_authentication"],
            get_two_factor_authentication_enabled(person),
        )


def person_two_factor_authentication_enabled(person):
    return (
        person["totp_enabled"]
        or person["email_otp_enabled"]
        or person["fido_enabled"]
    )


def person_two_factor_authentication_enabled_raw(person):
    return (
        person.totp_enabled or person.email_otp_enabled or person.fido_enabled
    )


def get_two_factor_authentication_enabled(person):
    two_factor_authentication_enabled = ["recovery_code"]
    if person["totp_enabled"]:
        two_factor_authentication_enabled.append("totp")
    if person["email_otp_enabled"]:
        two_factor_authentication_enabled.append("email_otp")
    if person["fido_enabled"]:
        two_factor_authentication_enabled.append("fido")
    return two_factor_authentication_enabled


def remove_otp_revovery_code(person_id, recovery_hash):
    """
    Remove an otp recovery code for a person_id.
    """
    person = Person.get(person_id)
    person.otp_recovery_codes.remove(recovery_hash.encode())
    flag_modified(person, "otp_recovery_codes")
    person.commit()
    persons_service.clear_person_cache()
    return person.serialize()


def check_totp(person, totp):
    """
    Check TOTP for a person.
    """
    return pyotp.TOTP(person["totp_secret"]).verify(totp)


def check_email_otp(person, email_otp):
    """
    Check email OTP for a person.
    """
    count = auth_tokens_store.get("email-otp-count-%s" % person["email"])
    if count is not None:
        if pyotp.HOTP(person["email_otp_secret"]).verify(
            email_otp, int(count)
        ):
            auth_tokens_store.delete("email-otp-count-%s" % person["email"])
            return True
    return False


def check_recovery_code(person, recovery_code):
    """
    Check recovery code for a person.
    """
    for recovery_hash in person["otp_recovery_codes"]:
        if flask_bcrypt.check_password_hash(recovery_hash, recovery_code):
            remove_otp_revovery_code(person["id"], recovery_hash)
            return True
    return False


def check_fido(person, authentication_response):
    """
    Check fido for a person.
    """
    try:
        state = session.pop("fido-state-%s" % person["id"])
    except KeyError:
        return False
    try:
        current_app.extensions["fido_server"].authenticate_complete(
            state,
            get_fido_attested_credential_data_from_person(
                person["fido_credentials"],
            ),
            authentication_response,
        )
    except BaseException:
        return False
    return True


def pre_enable_totp(person_id):
    """
    Pre-enable TOTP for a person.
    """
    person = Person.get(person_id)
    if person.totp_enabled:
        raise TOTPAlreadyEnabledException()
    else:
        person.totp_secret = pyotp.random_base32()
        totp = pyotp.TOTP(person.totp_secret)
        organisation = persons_service.get_organisation()
        totp_provisionning_uri = totp.provisioning_uri(
            name=person.email, issuer_name="Kitsu %s" % organisation["name"]
        )
        person.totp_enabled = False
        person.commit()
        persons_service.clear_person_cache()
        return totp_provisionning_uri, person.totp_secret


def enable_totp(person_id, totp):
    """
    Enable TOTP for a person.
    """
    person = Person.get(person_id)
    if person.totp_enabled:
        raise TOTPAlreadyEnabledException()
    elif check_totp(person.serialize(), totp):
        person.totp_enabled = True
        otp_recovery_codes = None
        if not person.otp_recovery_codes:
            otp_recovery_codes = generate_recovery_codes()
            person.otp_recovery_codes = hash_recovery_codes(otp_recovery_codes)
        if not person.preferred_two_factor_authentication:
            person.preferred_two_factor_authentication = "totp"
        person.commit()
        persons_service.clear_person_cache()
        return otp_recovery_codes
    else:
        raise WrongOTPException


def disable_totp(person_id):
    """
    Disable TOTP for a person.
    """
    person = Person.get(person_id)
    if not person.totp_enabled:
        raise TOTPNotEnabledException()
    person.totp_enabled = False
    person.totp_secret = None
    if not person_two_factor_authentication_enabled_raw(person):
        person.otp_recovery_codes = None
        person.preferred_two_factor_authentication = None
    elif person.preferred_two_factor_authentication == "totp":
        if person.fido_enabled:
            person.preferred_two_factor_authentication = "fido"
        elif person.email_otp_enabled:
            person.preferred_two_factor_authentication = "email_otp"
    person.commit()
    persons_service.clear_person_cache()
    return True


def pre_enable_email_otp(person_id):
    """
    Pre-enable mail OTP for a person.
    """
    person = Person.get(person_id)
    if person.email_otp_enabled:
        raise EmailOTPAlreadyEnabledException()
    else:
        person.email_otp_secret = pyotp.random_base32()
        person.email_otp_enabled = False
        person.commit()
        send_email_otp(person.serialize())
        persons_service.clear_person_cache()
        return True


def enable_email_otp(person_id, email_otp):
    """
    Enable mail OTP for a person.
    """
    person = Person.get(person_id)
    if person.email_otp_enabled:
        raise EmailOTPAlreadyEnabledException()
    elif check_email_otp(person.serialize(), email_otp):
        person.email_otp_enabled = True
        otp_recovery_codes = None
        if not person.otp_recovery_codes:
            otp_recovery_codes = generate_recovery_codes()
            person.otp_recovery_codes = hash_recovery_codes(otp_recovery_codes)
        if not person.preferred_two_factor_authentication:
            person.preferred_two_factor_authentication = "email_otp"
        person.commit()
        persons_service.clear_person_cache()
        return otp_recovery_codes
    else:
        raise WrongOTPException


def disable_email_otp(person_id):
    """
    Disable email OTP for a person.
    """
    person = Person.get(person_id)
    if not person.email_otp_enabled:
        raise EmailOTPNotEnabledException()
    person.email_otp_enabled = False
    person.email_otp_secret = None
    if not person_two_factor_authentication_enabled_raw(person):
        person.otp_recovery_codes = None
        person.preferred_two_factor_authentication = None
    elif person.preferred_two_factor_authentication == "email_otp":
        if person.fido_enabled:
            person.preferred_two_factor_authentication = "fido"
        elif person.totp_enabled:
            person.preferred_two_factor_authentication = "totp"
    person.commit()
    persons_service.clear_person_cache()
    return True


def send_email_otp(person):
    """
    Send an email with OTP and store the OTP for checking after.
    """
    count = random.randint(0, 999999999999)
    otp = pyotp.HOTP(person["email_otp_secret"]).at(count)
    auth_tokens_store.add(
        "email-otp-count-%s" % person["email"], count, ttl=60 * 5
    )
    organisation = persons_service.get_organisation()
    time_string = format_datetime(
        date_helpers.get_utc_now_datetime(),
        tzinfo=person["timezone"],
        locale=person["locale"],
    )
    person_IP = request.headers.get("X-Forwarded-For", None)
    html = f"""<p>Hello {person["first_name"]},</p>

<p>
Your verification code is : <strong>{otp}</strong>
</p>

<p>
This one time password will expire after 5 minutes. After, you will have to request a new one.
This email was sent at this date : {time_string}.
The IP of the person who requested this is: {person_IP}.
</p>

Thank you and see you soon on Kitsu,
</p>
<p>
{organisation["name"]} Team
</p>
"""
    subject = (
        f"{organisation['name']} - Kitsu : your verification code is {otp}"
    )
    emails.send_email(subject, html, person["email"])
    return True


def get_fido_attested_credential_data_from_person(
    fido_person_credentials=None,
):
    credentials = []
    if isinstance(fido_person_credentials, list):
        for credential in fido_person_credentials:
            credentials.append(
                AttestedCredentialData.create(
                    int2bytes(credential["aaguid"], 16),
                    int2bytes(credential["credential_id"]),
                    dict(
                        {
                            1: credential["public_key"]["1"],
                            3: credential["public_key"]["3"],
                            -1: credential["public_key"]["-1"],
                            -2: int2bytes(credential["public_key"]["-2"]),
                            -3: int2bytes(credential["public_key"]["-3"]),
                        }
                    ),
                )
            )
    return credentials


def pre_register_fido(person_id):
    """
    Pre-register FIDO device for a person.
    """
    person = Person.get(person_id)
    options, state = current_app.extensions["fido_server"].register_begin(
        PublicKeyCredentialUserEntity(
            id=str(person.id).encode(),
            name=person.email,
            display_name=person.full_name,
        ),
        credentials=get_fido_attested_credential_data_from_person(
            person.fido_credentials
        ),
        user_verification="preferred",
        authenticator_attachment="cross-platform",
    )
    session["fido-state-%s" % person.id] = state
    return dict(options.public_key)


def register_fido(person_id, registration_response, device_name):
    """
    Register FIDO device for a person.
    """
    person = Person.get(person_id)
    try:
        state = session.pop("fido-state-%s" % person.id)
    except KeyError:
        raise FIDONoPreregistrationException()
    try:
        auth_data = current_app.extensions["fido_server"].register_complete(
            state, registration_response
        )
    except BaseException:
        raise FIDOServerException()
    credential_data = {
        "device_name": device_name,
        "aaguid": bytes2int(auth_data.credential_data.aaguid),
        "credential_id": bytes2int(auth_data.credential_data.credential_id),
        "public_key": dict(
            {
                **auth_data.credential_data.public_key,
                -2: bytes2int(auth_data.credential_data.public_key.get(-2)),
                -3: bytes2int(auth_data.credential_data.public_key.get(-3)),
            }
        ),
    }
    if person.fido_credentials is None:
        person.fido_credentials = []
    person.fido_credentials.append(credential_data)
    flag_modified(person, "fido_credentials")
    person.fido_enabled = True
    otp_recovery_codes = None
    if not person.otp_recovery_codes:
        otp_recovery_codes = generate_recovery_codes()
        person.otp_recovery_codes = hash_recovery_codes(otp_recovery_codes)
    if not person.preferred_two_factor_authentication:
        person.preferred_two_factor_authentication = "fido"
    person.commit()
    persons_service.clear_person_cache()
    return otp_recovery_codes


def unregister_fido(person_id, device_name):
    """
    Unregister FIDO device for a person.
    """
    person = Person.get(person_id)
    for i in range(len(person.fido_credentials)):
        if person.fido_credentials[i]["device_name"] == device_name:
            del person.fido_credentials[i]
            break
    flag_modified(person, "fido_credentials")
    if len(person.fido_credentials) == 0:
        person.fido_enabled = False
    if not person_two_factor_authentication_enabled_raw(person):
        person.otp_recovery_codes = None
        person.preferred_two_factor_authentication = None
    elif person.preferred_two_factor_authentication == "fido":
        if person.totp_enabled:
            person.preferred_two_factor_authentication = "totp"
        elif person.email_otp_enabled:
            person.preferred_two_factor_authentication = "email_otp"
    person.commit()
    persons_service.clear_person_cache()
    return True


def get_challenge_fido(person_id):
    """
    Get new FIDO challenge for a person.
    """
    person = Person.get(person_id)
    options, state = current_app.extensions["fido_server"].authenticate_begin(
        credentials=get_fido_attested_credential_data_from_person(
            person.fido_credentials
        ),
    )
    session["fido-state-%s" % person.id] = state
    return dict(options.public_key)


def disable_two_factor_authentication_for_person(person_id):
    person = Person.get(person_id)
    if not person_two_factor_authentication_enabled_raw(person):
        raise TwoFactorAuthenticationNotEnabledException()
    person.email_otp_enabled = False
    person.email_otp_secret = None
    person.totp_enabled = False
    person.totp_secret = None
    person.fido_enabled = False
    person.fido_credentials = None
    person.otp_recovery_codes = None
    person.preferred_two_factor_authentication = None
    person.commit()
    persons_service.clear_person_cache()
    return True


def generate_new_recovery_codes(person_id):
    """
    Generate new recovery codes for a person.
    """
    person = Person.get(person_id)
    otp_recovery_codes = generate_recovery_codes()
    person.otp_recovery_codes = hash_recovery_codes(otp_recovery_codes)
    flag_modified(person, "otp_recovery_codes")
    person.commit()
    persons_service.clear_person_cache()
    return otp_recovery_codes


def revoke_tokens(app, jti):
    """
    Remove access and refresh tokens from auth token store.
    """
    auth_tokens_store.add(jti, "true", app.config["JWT_ACCESS_TOKEN_EXPIRES"])


def is_default_password(app, password):
    """
    Check if password is default.
    """
    return (
        password == "default"
        and app.config["AUTH_STRATEGY"] != "auth_local_no_password"
    )


def generate_reset_token():
    """
    Generate and return a reset token.
    """
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits)
        for _ in range(64)
    )


def generate_recovery_codes():
    """
    Generate and return recovery codes.
    """
    return [
        "".join(
            random.SystemRandom().choice(
                string.ascii_uppercase + string.digits
            )
            for _ in range(10)
        )
        for _ in range(16)
    ]


def hash_recovery_codes(recovery_codes):
    """
    Hash recovery codes given as argument and return them.
    """
    return [
        flask_bcrypt.generate_password_hash(recovery_code)
        for recovery_code in recovery_codes
    ]


def check_login_failed_attemps(person):
    """
    Check login failed attemps for a person.
    """
    login_failed_attemps = person["login_failed_attemps"]
    if login_failed_attemps is None:
        login_failed_attemps = 0
    if (
        login_failed_attemps >= 5
        and date_helpers.get_datetime_from_string(person["last_login_failed"])
        + timedelta(minutes=1)
        > date_helpers.get_utc_now_datetime()
    ):
        raise TooMuchLoginFailedAttemps()
    return login_failed_attemps


def logout(jti):
    try:
        revoke_tokens(current_app, jti)
    except Exception:
        pass



#!/usr/bin/env python
import base64
import orjson as json
import logging
import os
import sys
import tempfile
import zlib

from zou.remote.config_payload import (
    check_config_version,
    get_config_from_payload,
    get_storage,
    get_file_from_storage,
    put_file_to_storage,
    make_key,
)

from zou.utils.movie import (
    EncodingParameters,
    build_playlist_movie,
    concat_demuxer,
    concat_filter,
)

logging.basicConfig(
    format="%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)


def fetch_inputs(storage, outdir, preview_file_ids):
    """Fetch inputs from object storage, return a list of local paths"""
    input_paths = []
    for input_id in preview_file_ids:
        filename = f"cache-previews-{input_id}.mp4"
        file_path = os.path.join(outdir, filename)
        input_paths.append(
            (
                get_file_from_storage(
                    storage, file_path, make_key("previews", input_id)
                ),
                filename,
            )
        )
    return input_paths


def _run_build_playlist(input_paths, output_movie_path, enc_params, full):
    is_build_successful = False
    if not full:
        try:
            result = build_playlist_movie(
                concat_demuxer,
                input_paths,
                output_movie_path,
                **enc_params._asdict(),
            )
            if result["success"] and os.path.exists(output_movie_path):
                is_build_successful = True
        except Exception:
            is_build_successful = False

    if not is_build_successful:
        result = build_playlist_movie(
            concat_filter,
            input_paths,
            output_movie_path,
            **enc_params._asdict(),
        )
    return result


def main():
    """
    Generate a playlist from shots, upload generated file in the object
    storage.
    We assume that this script is launched in the context of a Nomad job.
    Arguments are retrieved from a payload file at JSON format.
    Nevertheless this script can be run directly from the command line.
    """
    config = get_config_from_payload()
    check_config_version(config)
    storage = get_storage(config, "movies")

    with tempfile.TemporaryDirectory() as tmpdir:
        enc_params = EncodingParameters(
            width=config["width"], height=config["height"], fps=config["fps"]
        )

        input_zipped = base64.b64decode(config["input"])
        preview_file_ids = json.loads(zlib.decompress(input_zipped))
        input_paths = fetch_inputs(storage, tmpdir, preview_file_ids)

        output_movie_path = os.path.join(tmpdir, config["output_filename"])
        result = _run_build_playlist(
            input_paths,
            output_movie_path,
            enc_params,
            config["full"] == "true",
        )

        if result["success"]:
            put_file_to_storage(
                storage, output_movie_path, config["output_key"]
            )
        else:
            logger.error(
                "Playlist creation failed: %s" % result.get("message"),
            )
            sys.exit(1)


if __name__ == "__main__":
    main()



import orjson as json

from flask import abort, request, send_file as flask_send_file, current_app
from flask_restful import Resource, reqparse
from flask_jwt_extended import jwt_required

from zou.app.mixin import ArgsMixin
from zou.app.utils import permissions, date_helpers

from zou.app.services import (
    chats_service,
    comments_service,
    deletion_service,
    entities_service,
    persons_service,
    tasks_service,
    user_service,
)
from zou.app import config


class DownloadAttachmentResource(Resource):
    @jwt_required()
    def get(self, attachment_file_id, file_name):
        """
        Download attachment file.
        ---
        tags:
        - Comments
        produces:
          - multipart/form-data
          - image/png
          - image/gif
          - image/jpeg
        parameters:
          - in: path
            name: attachment_file_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: file_name
            required: True
            type: string
            x-example: filename
        responses:
            200:
                description: Attachment file downloaded
                schema:
                    type: file
            404:
                description: Download failed
        """
        attachment_file = comments_service.get_attachment_file(
            attachment_file_id
        )
        if attachment_file["comment_id"] is not None:
            comment = tasks_service.get_comment(attachment_file["comment_id"])
            user_service.check_task_access(comment["object_id"])
        elif attachment_file["chat_message_id"] is not None:
            message = chats_service.get_chat_message(
                attachment_file["chat_message_id"]
            )
            chat = chats_service.get_chat_by_id(message["chat_id"])
            entity = entities_service.get_entity(chat["object_id"])
            user_service.check_project_access(entity["project_id"])
        else:
            raise permissions.PermissionDenied()
        try:
            file_path = comments_service.get_attachment_file_path(
                attachment_file
            )
            return flask_send_file(
                file_path,
                conditional=True,
                mimetype=attachment_file["mimetype"],
                as_attachment=False,
                download_name=attachment_file["name"],
                max_age=config.CLIENT_CACHE_MAX_AGE,
                last_modified=date_helpers.get_datetime_from_string(
                    attachment_file["updated_at"]
                ),
            )
        except Exception:
            if config.LOG_FILE_NOT_FOUND:
                current_app.logger.error(
                    f"Attachment file was not found for: {attachment_file_id}"
                )
            abort(404)


class AckCommentResource(Resource):
    """
    Acknowledge given comment. If it's already acknowledged, remove
    acknowledgement.
    """

    @jwt_required()
    def post(self, task_id, comment_id):
        """
        Acknowledge given comment.
        ---
        tags:
        - Comments
        description: If it's already acknowledged, remove acknowledgement.
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: comment_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Comment acknowledged
        """
        user_service.check_task_access(task_id)
        return comments_service.acknowledge_comment(comment_id)


class CommentTaskResource(Resource):
    """
    Creates a new comment for given task. It requires a text, a task_status
    and a person as arguments. This way, comments keep history of status
    changes. When the comment is created, it updates the task status with
    given task status.
    """

    @jwt_required()
    def post(self, task_id):
        """
        Create a new comment for given task.
        ---
        tags:
        - Comments
        description: It requires a text, a task_status and a person as arguments.
                     This way, comments keep history of status changes.
                     When the comment is created, it updates the task status with given task status.
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: body
            name: Comment
            description: person ID, name, comment, revision and change status of task
            schema:
                type: object
                required:
                    - task_status_id
                properties:
                    task_status_id:
                        type: string
                        format: UUID
                        example: a24a6ea4-ce75-4665-a070-57453082c25a4-ce75-4665-a070-57453082c25
                    comment:
                        type: string
                    person_id:
                        type: string
                        format: UUID
                        example: a24a6ea4-ce75-4665-a070-57453082c25a4-ce75-4665-a070-57453082c25
                    created_at:
                        type: string
                        format: date-time
                        example: "2022-07-12T13:00:00"
                    checklist:
                        type: object
                        properties:
                            item 1:
                                type: string
        responses:
            201:
                description: New comment created
        """
        (
            task_status_id,
            comment,
            person_id,
            created_at,
            checklist,
            links,
        ) = self.get_arguments()

        user_service.check_task_action_access(task_id)
        user_service.check_task_status_access(task_status_id)
        files = request.files

        if not permissions.has_manager_permissions():
            person_id = None
            created_at = None
        comment = comments_service.create_comment(
            person_id,
            task_id,
            task_status_id,
            comment,
            checklist,
            files,
            created_at,
            links,
        )
        return comment, 201

    def get_arguments(self):
        parser = reqparse.RequestParser()
        if request.is_json:
            location = ["values", "json"]
            parser.add_argument(
                "checklist",
                type=dict,
                action="append",
                default=[],
                location=location,
            )
            parser.add_argument(
                "links",
                type=str,
                action="append",
                default=[],
                location=location,
            )
        else:
            location = "values"
            parser.add_argument("checklist", default="[]", location=location)
            parser.add_argument("links", default="[]", location=location)
        parser.add_argument(
            "task_status_id",
            required=True,
            help="Task Status ID is missing",
            location=location,
        )
        parser.add_argument("comment", default="", location=location)
        parser.add_argument("person_id", default="", location=location)
        parser.add_argument("created_at", default="", location=location)
        args = parser.parse_args()
        return (
            args["task_status_id"],
            args["comment"],
            args["person_id"],
            args["created_at"],
            (
                args["checklist"]
                if request.is_json
                else json.loads(args["checklist"])
            ),
            (args["links"] if request.is_json else json.loads(args["links"])),
        )


class AttachmentResource(Resource):
    @jwt_required()
    def delete(self, task_id, comment_id, attachment_id):
        """
        Delete attachment linked to a comment matching given ID.
        ---
        tags:
        - Comments
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: comment_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: attachment_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Empty response
        """
        user = persons_service.get_current_user()
        comment = tasks_service.get_comment(comment_id)
        if comment["person_id"] != user["id"]:
            task = tasks_service.get_task(task_id)
            user_service.check_manager_project_access(task["project_id"])

        deletion_service.remove_attachment_file_by_id(attachment_id)
        return "", 204


class AddAttachmentToCommentResource(Resource):
    @jwt_required()
    def post(self, task_id, comment_id):
        """
        Add given files to the comment entry as attachments.
        ---
        tags:
        - Comments
        consumes:
          - image/png
          - image/gif
          - image/jpeg
          - multipart/form-data
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: comment_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: formData
            name: files
            type: file
            required: True
        responses:
            201:
                description: Given files added to the comment entry as attachments
        """
        user = persons_service.get_current_user()
        comment = tasks_service.get_comment(comment_id)
        if comment["person_id"] != user["id"]:
            task = tasks_service.get_task(task_id)
            user_service.check_manager_project_access(task["project_id"])

        files = request.files
        comment = comments_service.add_attachments_to_comment(comment, files)
        return comment["attachment_files"], 201


class CommentManyTasksResource(Resource):
    """
    Create several comments at once. Each comment requires a text, a task id,
    a task_status and a person as arguments. This way, comments keep history of
    status changes. When the comment is created, it updates the task status with
    given task status.
    """

    @jwt_required()
    def post(self, project_id):
        """
        Create several comments at once.
        ---
        tags:
        - Comments
        description: Each comment requires a text, a task id, a task_status and a person as arguments.
                     This way, comments keep history of status changes.
                     When the comment is created, it updates the task status with given task status.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: body
            name: Comment
            description: person ID, name, comment, revision and change status of task
            schema:
                type: object
                required:
                    - task_status_id
                properties:
                    task_status_id:
                        type: string
                        format: UUID
                        example: a24a6ea4-ce75-4665-a070-57453082c25a4-ce75-4665-a070-57453082c25
                    comment:
                        type: string
                    person_id:
                        type: string
                        format: UUID
                        example: a24a6ea4-ce75-4665-a070-57453082c25a4-ce75-4665-a070-57453082c25
                    object_id:
                        type: string
                        format: UUID
                        example: a24a6ea4-ce75-4665-a070-57453082c25a4-ce75-4665-a070-57453082c25
                    created_at:
                        type: string
                        format: date-time
                        example: "2022-07-12T13:00:00"
                    checklist:
                        type: object
                        properties:
                            item 1:
                                type: string
        responses:
            201:
                description: Given files added to the comment entry as attachments
        """
        comments = request.json
        person = persons_service.get_current_user(relations=True)
        try:
            user_service.check_manager_project_access(project_id)
        except permissions.PermissionDenied:
            comments = self.get_allowed_comments_only(comments, person)
        result = []
        for comment in comments:
            try:
                user_service.check_task_status_access(
                    comment["task_status_id"]
                )
                comment = comments_service.create_comment(
                    person["id"],
                    comment["object_id"],
                    comment["task_status_id"],
                    comment["comment"],
                    [],
                    {},
                    None,
                    comment.get("links", []),
                )
                result.append(comment)
            except KeyError:
                pass
        return result, 201

    def get_allowed_comments_only(self, comments, person):
        allowed_comments = []
        for comment in comments:
            try:
                task = tasks_service.get_task(
                    comment["object_id"], relations=True
                )
                if (
                    person["role"] == "supervisor"
                    and (
                        len(person["departments"]) == 0
                        or tasks_service.get_task_type(task["task_type_id"])[
                            "department_id"
                        ]
                        in person["departments"]
                    )
                ) or person["id"] in task["assignees"]:
                    allowed_comments.append(comment)
            except permissions.PermissionDenied:
                pass
            except KeyError:
                pass
        return allowed_comments


class ReplyCommentResource(Resource, ArgsMixin):
    """
    Reply to given comment. Add comment to its replies list.
    """

    @jwt_required()
    def post(self, task_id, comment_id):
        """
        Reply to given comment.
        ---
        tags:
        - Comments
        description: Add comment to its replies list.
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: comment_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: formData
            name: text
            type: string
            x-example: comment
        responses:
            200:
                description: Reply to given comment
        """
        args = self.get_args(
            [
                ("text", "", False),
            ]
        )

        comment = tasks_service.get_comment(comment_id)
        current_user = persons_service.get_current_user()
        if comment["person_id"] != current_user["id"]:
            if permissions.has_client_permissions():
                raise permissions.PermissionDenied()
            user_service.check_task_action_access(task_id)

        return comments_service.reply_comment(comment_id, args["text"])


class DeleteReplyCommentResource(Resource):
    """
    Delete given comment reply.
    """

    @jwt_required()
    def delete(self, task_id, comment_id, reply_id):
        """
        Delete given comment reply.
        ---
        tags:
        - Comments
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: comment_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: reply_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Given comment reply deleted
        """
        reply = comments_service.get_reply(comment_id, reply_id)
        current_user = persons_service.get_current_user()
        if reply["person_id"] != current_user["id"]:
            permissions.check_admin_permissions()
        return comments_service.delete_reply(comment_id, reply_id)


class ProjectAttachmentFiles(Resource):
    @jwt_required()
    def get(self, project_id):
        """
        Return all attachment files related to given project.
        ---
        tags:
        - Comments
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All attachment files related to given project
        """
        permissions.check_admin_permissions()
        return comments_service.get_all_attachment_files_for_project(
            project_id
        )


class TaskAttachmentFiles(Resource):
    """
    Return all attachment files related to given task.
    """

    @jwt_required()
    def get(self, task_id):
        """
        Return all attachment files related to given task.
        ---
        tags:
        - Comments
        parameters:
          - in: path
            name: task_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All attachment files related to given task
        """
        permissions.check_admin_permissions()
        return comments_service.get_all_attachment_files_for_task(task_id)



from sqlalchemy_utils import UUIDType, IPAddressType, ChoiceType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin

ORIGINS = [("web", "Web"), ("script", "Script")]


class LoginLog(db.Model, BaseMixin, SerializerMixin):
    """
    Table to log all web session logins. The aim is to build a table that
    helps finding suspicious behaviours.
    """

    person_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        nullable=False,
        index=True,
    )
    ip_address = db.Column(IPAddressType)
    origin = db.Column(ChoiceType(ORIGINS))



from zou.app.services import shots_service
from zou.app.utils import events


def add_shot_to_scene(scene, shot):
    """
    Link a shot to a scene. Once done, the scene is considered as the source
    of the scene.
    """
    shot = shots_service.update_shot(shot["id"], {"source_id": scene["id"]})
    events.emit(
        "shot:add-to-scene",
        {"scene_id": scene["id"], "shot_id": shot["id"]},
        project_id=shot["project_id"],
    )
    return shot


def remove_shot_from_scene(scene, shot):
    """
    Remove link from a shot to a scene.
    """
    shot = shots_service.update_shot(shot["id"], {"source_id": None})
    events.emit(
        "shot:remove-from-scene",
        {"scene_id": scene["id"], "shot_id": shot["id"]},
        project_id=shot["project_id"],
    )
    return shot


def get_shots_by_scene(scene_id):
    """
    Get shots linked to a scene (of which the source is the given scene).
    """
    return shots_service.get_shots({"source_id": scene_id})
